from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.constants import UNIQUE_CONFLICT_METHOD_UPDATE
from mage_ai.io.export_utils import BadConversionError, PandasTypes
from mage_ai.io.sql import BaseSQL
from mage_ai.shared.parsers import encode_complex
from mage_ai.shared.utils import is_port_in_use
from pandas import DataFrame, Series
from psycopg2 import connect, _psycopg
from sshtunnel import SSHTunnelForwarder
from typing import IO, List, Union
import numpy as np
import pandas as pd
import simplejson


class Postgres(BaseSQL):
    """
    Handles data transfer between a PostgreSQL database and the Mage app.
    """
    def __init__(
        self,
        dbname: str,
        user: str,
        password: str,
        host: str,
        port: Union[str, None] = None,
        schema: str = None,
        connection_method: str = 'direct',
        ssh_host: Union[str, None] = None,
        ssh_port: Union[str, None] = None,
        ssh_username: Union[str, None] = None,
        ssh_password: Union[str, None] = None,
        ssh_pkey: Union[str, None] = None,
        verbose=True,
        connect_timeout: int = None,
        **kwargs,
    ) -> None:
        """
        Initializes the data loader.

        Args:
            dbname (str): The name of the database to connect to.
            user (str): The user with which to connect to the database with.
            password (str): The login password for the user.
            host (str): Path to host address for database.
            port (str): Port on which the database is running.
            **kwargs: Additional settings for creating SQLAlchemy engine and connection
        """
        self.ssh_tunnel = None
        super().__init__(
            verbose=verbose,
            dbname=dbname,
            user=user,
            password=password,
            host=host,
            port=port,
            schema=schema,
            connection_method=connection_method,
            ssh_host=ssh_host,
            ssh_port=ssh_port,
            ssh_username=ssh_username,
            ssh_password=ssh_password,
            ssh_pkey=ssh_pkey,
            connect_timeout=connect_timeout,
            **kwargs,
        )

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'Postgres':
        return cls(
            dbname=config[ConfigKey.POSTGRES_DBNAME],
            user=config[ConfigKey.POSTGRES_USER],
            password=config[ConfigKey.POSTGRES_PASSWORD],
            host=config[ConfigKey.POSTGRES_HOST],
            port=config[ConfigKey.POSTGRES_PORT],
            schema=config[ConfigKey.POSTGRES_SCHEMA],
            connection_method=config[ConfigKey.POSTGRES_CONNECTION_METHOD],
            ssh_host=config[ConfigKey.POSTGRES_SSH_HOST],
            ssh_port=config[ConfigKey.POSTGRES_SSH_PORT],
            ssh_username=config[ConfigKey.POSTGRES_SSH_USERNAME],
            ssh_password=config[ConfigKey.POSTGRES_SSH_PASSWORD],
            ssh_pkey=config[ConfigKey.POSTGRES_SSH_PKEY],
            connect_timeout=config[ConfigKey.POSTGRES_CONNECT_TIMEOUT],
        )

    def default_database(self) -> str:
        return self.settings['dbname']

    def default_schema(self) -> str:
        return self.settings.get('schema')

    def open(self) -> None:
        with self.printer.print_msg('Opening connection to PostgreSQL database'):
            database = self.settings['dbname']
            host = self.settings['host']
            password = self.settings['password']
            port = self.settings['port']
            user = self.settings['user']
            if self.settings['connection_method'] == 'ssh_tunnel':
                ssh_setting = dict(ssh_username=self.settings['ssh_username'])
                if self.settings['ssh_pkey'] is not None:
                    ssh_setting['ssh_pkey'] = self.settings['ssh_pkey']
                else:
                    ssh_setting['ssh_password'] = self.settings['ssh_password']

                # Find an available local port
                local_port = port
                max_local_port = local_port + 100
                while is_port_in_use(local_port):
                    if local_port > max_local_port:
                        raise Exception(
                            'Unable to find an open port, please clear your running processes '
                            'if possible.'
                        )
                    local_port += 1
                self.ssh_tunnel = SSHTunnelForwarder(
                    (self.settings['ssh_host'], self.settings['ssh_port']),
                    remote_bind_address=(host, port),
                    local_bind_address=('', local_port),
                    **ssh_setting,
                )
                self.ssh_tunnel.start()
                self.ssh_tunnel._check_is_started()

                host = '127.0.0.1'
                port = self.ssh_tunnel.local_bind_port

            connect_opts = dict(
                database=database,
                host=host,
                password=password,
                port=port,
                user=user,
                keepalives=1,
                keepalives_idle=300,
            )

            if self.settings.get('connect_timeout'):
                connect_opts['connect_timeout'] = self.settings['connect_timeout']

            try:
                self._ctx = connect(**connect_opts)
            except Exception:
                if self.ssh_tunnel is not None:
                    self.ssh_tunnel.stop()
                    self.ssh_tunnel = None

    def close(self) -> None:
        """
        Close the underlying connection to the SQL data source if open. Else will do nothing.
        """
        if '_ctx' in self.__dict__:
            self._ctx.close()
            del self._ctx
        if self.ssh_tunnel is not None:
            self.ssh_tunnel.stop()
            self.ssh_tunnel = None
        if self.verbose and self.printer.exists_previous_message:
            print('')

    def table_exists(self, schema_name: str, table_name: str) -> bool:
        with self.conn.cursor() as cur:
            cur.execute(
                f'SELECT * FROM pg_tables WHERE schemaname = \'{schema_name}\' AND '
                f'tablename = \'{table_name}\''
            )
            return bool(cur.rowcount)

    def get_type(self, column: Series, dtype: str) -> str:
        if dtype in (
            PandasTypes.MIXED,
            PandasTypes.UNKNOWN_ARRAY,
            PandasTypes.COMPLEX,
        ):
            series = column[column.notnull()]
            values = series.values

            column_type = None

            if len(values) >= 1:
                column_type = 'JSONB'

                values_not_empty_list = [v for v in values if type(v) is not list or v]
                if not values_not_empty_list:
                    # All values are empty list
                    return column_type
                value = values_not_empty_list[0]
                if type(value) is list:
                    if len(value) >= 1:
                        item = value[0]
                        if type(item) is dict:
                            column_type = 'JSONB'
                        else:
                            item_series = pd.Series(data=item)
                            item_dtype = item_series.dtype
                            if PandasTypes.OBJECT != item_dtype:
                                item_type = self.get_type(item_series, item_dtype)
                                column_type = f'{item_type}[]'
                            else:
                                column_type = 'text[]'
                    else:
                        column_type = 'text[]'

            if column_type:
                return column_type

            raise BadConversionError(
                f'Cannot convert column \'{column.name}\' with data type \'{dtype}\' to '
                'a PostgreSQL datatype.'
            )
        elif dtype in (PandasTypes.DATETIME, PandasTypes.DATETIME64):
            try:
                if column.dt.tz:
                    return 'timestamptz'
            except AttributeError:
                pass
            return 'timestamp'
        elif dtype == PandasTypes.TIME:
            try:
                if column.dt.tz:
                    return 'timetz'
            except AttributeError:
                pass
            return 'time'
        elif dtype == PandasTypes.DATE:
            return 'date'
        elif dtype == PandasTypes.STRING:
            return 'text'
        elif dtype == PandasTypes.CATEGORICAL:
            return 'text'
        elif dtype == PandasTypes.BYTES:
            return 'bytea'
        elif dtype in (PandasTypes.FLOATING, PandasTypes.DECIMAL, PandasTypes.MIXED_INTEGER_FLOAT):
            return 'double precision'
        elif dtype == PandasTypes.INTEGER or dtype == PandasTypes.INT64:
            max_int, min_int = column.max(), column.min()
            if np.int16(max_int) == max_int and np.int16(min_int) == min_int:
                return 'smallint'
            elif np.int32(max_int) == max_int and np.int32(min_int) == min_int:
                return 'integer'
            else:
                return 'bigint'
        elif dtype == PandasTypes.BOOLEAN:
            return 'boolean'
        elif dtype in (PandasTypes.TIMEDELTA, PandasTypes.TIMEDELTA64, PandasTypes.PERIOD):
            return 'bigint'
        elif dtype == PandasTypes.EMPTY:
            return 'text'
        elif PandasTypes.OBJECT == dtype:
            return 'JSONB'
        else:
            print(f'Invalid datatype provided: {dtype}')

        return 'text'

    def upload_dataframe(
        self,
        cursor: _psycopg.cursor,
        df: DataFrame,
        db_dtypes: List[str],
        dtypes: List[str],
        full_table_name: str,
        buffer: Union[IO, None] = None,
        allow_reserved_words: bool = False,
        unique_conflict_method: str = None,
        unique_constraints: List[str] = None,
    ) -> None:
        if unique_constraints and unique_conflict_method:
            use_insert_command = True
        else:
            # Use COPY command
            use_insert_command = False

        def clean_array_value(val):
            if val is None or type(val) is not str or len(val) < 2:
                return val
            if val[0] == '[' and val[-1] == ']':
                return '{' + val[1:-1] + '}'
            return val

        def serialize_obj(val):
            if type(val) is dict:
                return simplejson.dumps(
                    val,
                    default=encode_complex,
                    ignore_nan=True,
                )
            elif type(val) is list and len(val) >= 1 and type(val[0]) is dict:
                return simplejson.dumps(
                    val,
                    default=encode_complex,
                    ignore_nan=True,
                )
            elif not use_insert_command and type(val) is list:
                return clean_array_value(simplejson.dumps(
                    val,
                    default=encode_complex,
                    ignore_nan=True,
                ))
            return val

        df_ = df.copy()
        columns = df_.columns

        for col in columns:
            df_col_dropna = df_[col].dropna()
            if df_col_dropna.count() == 0:
                continue
            if dtypes[col] == PandasTypes.OBJECT \
                    or (df_[col].dtype == PandasTypes.OBJECT and
                        type(df_col_dropna.iloc[0]) != str):
                df_[col] = df_[col].apply(lambda x: serialize_obj(x))
        df_.replace({np.NaN: None}, inplace=True)

        insert_columns = ', '.join([f'"{col}"'for col in columns])

        if use_insert_command:
            # Use INSERT command
            values_placeholder = ', '.join(["%s" for i in range(len(columns))])
            values = []
            for i, row in df_.iterrows():
                values.append(tuple(row))
            commands = [
                f'INSERT INTO {full_table_name} ({insert_columns})',
                f'VALUES ({values_placeholder})',
            ]

            unique_constraints = \
                [f'"{self._clean_column_name(col, allow_reserved_words=allow_reserved_words)}"'
                 for col in unique_constraints]
            columns_cleaned = \
                [f'"{self._clean_column_name(col, allow_reserved_words=allow_reserved_words)}"'
                 for col in columns]

            commands.append(f"ON CONFLICT ({', '.join(unique_constraints)})")
            if UNIQUE_CONFLICT_METHOD_UPDATE == unique_conflict_method:
                update_command = [f'{col} = EXCLUDED.{col}' for col in columns_cleaned]
                commands.append(
                    f"DO UPDATE SET {', '.join(update_command)}",
                )
            else:
                commands.append('DO NOTHING')
            cursor.executemany('\n'.join(commands), values)
        else:
            # Use COPY command
            df_.to_csv(
                buffer,
                header=False,
                index=False,
                na_rep='',
            )
            buffer.seek(0)
            cursor.copy_expert(f"""
COPY {full_table_name} ({insert_columns}) FROM STDIN (
    FORMAT csv
    , DELIMITER \',\'
    , NULL \'\'
    , FORCE_NULL({insert_columns})
);
        """, buffer)



"""Add unique indexes on authentication tables

Revision ID: 643b6e65e814
Revises: 97ff9f55a3c0
Create Date: 2023-01-31 22:04:43.403008

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '643b6e65e814'
down_revision = '97ff9f55a3c0'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('oauth2_access_token', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_oauth2_access_token_token'), ['token'], unique=True)

    with op.batch_alter_table('oauth2_application', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_oauth2_application_client_id'), ['client_id'], unique=True)

    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_user_email'), ['email'], unique=True)
        batch_op.create_index(batch_op.f('ix_user_username'), ['username'], unique=True)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_user_username'))
        batch_op.drop_index(batch_op.f('ix_user_email'))

    with op.batch_alter_table('oauth2_application', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_oauth2_application_client_id'))

    with op.batch_alter_table('oauth2_access_token', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_oauth2_access_token_token'))

    # ### end Alembic commands ###



from mage_ai.data_preparation.models.block import PYTHON_COMMAND
from mage_ai.server.logger import Logger
from typing import List, Dict
import importlib
import json
import subprocess
import traceback

logger = Logger().new_server_logger(__name__)


def get_collection(key: str, available_options: List[Dict]):
    collection = []

    for option in available_options:
        d = option.copy()
        if not d.get('uuid'):
            d['uuid'] = d['name'].lower().replace(' ', '_')
        module_name = d.get('module_name', d['name'].replace(' ', ''))
        uuid = d['uuid']
        try:
            module = importlib.import_module(f"mage_integrations.{key}.{uuid}")
            mod = getattr(module, module_name)
            d['templates'] = mod.templates()
        except FileNotFoundError:
            d['templates'] = {}
        except Exception:
            try:
                absolute_file_path = '/'.join(module.__file__.split('/')[:-2])
                absolute_file_path = f'{absolute_file_path}/{uuid}/__init__.py'
                proc = subprocess.run([
                    PYTHON_COMMAND,
                    absolute_file_path,
                    '--config_json',
                    json.dumps({}),
                    '--show_templates',
                ], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                for line in proc.stdout.decode().split('\n'):
                    try:
                        d['templates'] = json.loads(line)
                    except Exception:
                        pass
            except Exception as err:
                logger.error(f"Failed to load source {d['uuid']}: {err}")
                logger.error(traceback.format_exc())
                continue

        collection.append(d)

    return collection



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.monitor.monitor_stats import MonitorStats


class MonitorStatResource(GenericResource):
    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        query = kwargs.get('query', {})

        pipeline_uuids = query.get('pipeline_uuid', None)
        if pipeline_uuids:
            pipeline_uuid = pipeline_uuids[0]
        else:
            pipeline_uuid = None

        start_times = query.get('start_time', None)
        if start_times:
            start_time = start_times[0]
        else:
            start_time = None

        end_times = query.get('end_time', None)
        if end_times:
            end_time = end_times[0]
        else:
            end_time = None

        pipeline_schedule_ids = query.get('pipeline_schedule_id', None)
        if pipeline_schedule_ids:
            pipeline_schedule_id = pipeline_schedule_ids[0]
        else:
            pipeline_schedule_id = None

        stats = MonitorStats().get_stats(
            pk,
            pipeline_uuid=pipeline_uuid,
            start_time=start_time,
            end_time=end_time,
            pipeline_schedule_id=pipeline_schedule_id,
        )

        return self(dict(
            stats_type=pk,
            stats=stats,
        ), user, **kwargs)



"""Add indices to schedule models.

Revision ID: 1bfc6d904929
Revises: 66e67039b8a2
Create Date: 2023-05-23 15:59:05.328359

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1bfc6d904929'
down_revision = '66e67039b8a2'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('block_run', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_block_run_pipeline_run_id'), ['pipeline_run_id'], unique=False)

    with op.batch_alter_table('pipeline_run', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_pipeline_run_backfill_id'), ['backfill_id'], unique=False)
        batch_op.create_index(batch_op.f('ix_pipeline_run_pipeline_schedule_id'), ['pipeline_schedule_id'], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('pipeline_run', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_pipeline_run_pipeline_schedule_id'))
        batch_op.drop_index(batch_op.f('ix_pipeline_run_backfill_id'))

    with op.batch_alter_table('block_run', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_block_run_pipeline_run_id'))

    # ### end Alembic commands ###



import os
from datetime import datetime

from freezegun import freeze_time

from mage_ai.data_preparation.models.constants import PipelineType
from mage_ai.data_preparation.models.triggers import ScheduleStatus
from mage_ai.data_preparation.repo_manager import get_repo_config
from mage_ai.orchestration.db.models.schedules import PipelineRun, PipelineSchedule
from mage_ai.orchestration.pipeline_scheduler import configure_pipeline_run_payload
from mage_ai.tests.base_test import DBTestCase
from mage_ai.tests.factory import (
    create_pipeline_run,
    create_pipeline_run_with_schedule,
    create_pipeline_with_blocks,
)


class PipelineScheduleTests(DBTestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()
        self.pipeline = create_pipeline_with_blocks(
            'test pipeline',
            self.repo_path,
        )

    def test_pipeline_runs_count(self):
        pipeline_schedule = PipelineSchedule.create(pipeline_uuid='test_pipeline')
        for _ in range(5):
            create_pipeline_run_with_schedule(
                'test_pipeline',
                pipeline_schedule_id=pipeline_schedule.id,
            )
        self.assertEqual(pipeline_schedule.pipeline_runs_count, 5)

    def test_validate_schedule_interval(self):
        PipelineSchedule.create(
            pipeline_uuid='test_pipeline',
            schedule_interval='@daily'
        )
        PipelineSchedule.create(
            pipeline_uuid='test_pipeline',
            schedule_interval='* * * * *'
        )
        with self.assertRaises(ValueError) as context:
            PipelineSchedule.create(
                pipeline_uuid='test_pipeline',
                schedule_interval='random_str'
            )
        self.assertTrue('Cron expression is invalid.' in str(context.exception))

    def test_active_schedules(self):
        create_pipeline_with_blocks(
            'test active schedule 1',
            self.repo_path,
        )
        create_pipeline_with_blocks(
            'test active schedule 2',
            self.repo_path,
        )
        PipelineSchedule.create(
            pipeline_uuid='test_active_schedule_1',
            schedule_interval='@daily'
        )
        pipeline_schedule2 = PipelineSchedule.create(
            pipeline_uuid='test_active_schedule_1',
            schedule_interval='@daily'
        )
        pipeline_schedule3 = PipelineSchedule.create(
            pipeline_uuid='test_active_schedule_2',
            schedule_interval='@daily'
        )
        PipelineSchedule.create(
            pipeline_uuid='test_active_schedule_2',
            schedule_interval='@daily'
        )
        pipeline_schedule2.update(status=ScheduleStatus.ACTIVE)
        pipeline_schedule3.update(status=ScheduleStatus.ACTIVE)
        results1 = PipelineSchedule.active_schedules(pipeline_uuids=['test_active_schedule_1'])
        results2 = PipelineSchedule.active_schedules(pipeline_uuids=['test_active_schedule_2'])
        results3 = PipelineSchedule.active_schedules()
        self.assertEqual(set([r.id for r in results1]), set([pipeline_schedule2.id]))
        self.assertEqual(set([r.id for r in results2]), set([pipeline_schedule3.id]))
        self.assertEqual(
            set([r.id for r in results3]),
            set([pipeline_schedule2.id, pipeline_schedule3.id]),
        )

    def test_should_schedule(self):
        pipeline_schedule1 = PipelineSchedule.create(
            pipeline_uuid='test_pipeline',
            schedule_interval='@daily'
        )
        pipeline_schedule2 = PipelineSchedule.create(
            pipeline_uuid='test_pipeline',
            schedule_interval='@daily'
        )
        pipeline_schedule2.update(status=ScheduleStatus.ACTIVE)
        self.assertFalse(pipeline_schedule1.should_schedule())
        self.assertTrue(pipeline_schedule2.should_schedule())


class PipelineRunTests(DBTestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()
        self.pipeline = create_pipeline_with_blocks(
            'test pipeline',
            self.repo_path,
        )

    def test_block_runs_count(self):
        pipeline_run = create_pipeline_run(pipeline_uuid='test_pipeline')
        block_count = len(self.__class__.pipeline.get_executable_blocks())
        self.assertEqual(pipeline_run.block_runs_count, block_count)

    def test_execution_partition(self):
        execution_date = datetime.now()
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_pipeline',
            execution_date=execution_date,
        )
        execution_date_str = execution_date.strftime(format='%Y%m%dT%H%M%S')
        self.assertEqual(
            pipeline_run.execution_partition,
            f'{pipeline_run.pipeline_schedule_id}/{execution_date_str}',
        )

    @freeze_time('2023-05-01 01:20:33')
    def test_execution_partition_from_variables(self):
        pipeline_schedule = PipelineSchedule.create(pipeline_uuid='test_pipeline')
        payload = configure_pipeline_run_payload(pipeline_schedule, PipelineType.PYTHON, dict())[0]
        pipeline_run = PipelineRun.create(**payload)
        execution_date_str = datetime.utcnow().strftime(format='%Y%m%dT%H%M%S_%f')
        self.assertEqual(
            pipeline_run.execution_partition,
            f'{pipeline_run.pipeline_schedule_id}/{execution_date_str}',
        )

    def test_log_file(self):
        execution_date = datetime.now()
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_pipeline',
            execution_date=execution_date,
        )
        execution_date_str = execution_date.strftime(format='%Y%m%dT%H%M%S')
        expected_file_path = os.path.join(
            get_repo_config(self.repo_path).variables_dir,
            'pipelines/test_pipeline/.logs',
            f'{pipeline_run.pipeline_schedule_id}/{execution_date_str}/pipeline.log',
        )
        self.assertEqual(pipeline_run.logs.get('path'), expected_file_path)

    def test_active_runs(self):
        create_pipeline_with_blocks(
            'test active run 1',
            self.repo_path,
        )
        create_pipeline_with_blocks(
            'test active run 2',
            self.repo_path,
        )
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_active_run_1',
        )
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        pipeline_schedule = pipeline_run.pipeline_schedule
        pipeline_run2 = create_pipeline_run_with_schedule(
            pipeline_uuid='test_active_run_1',
            pipeline_schedule_id=pipeline_schedule.id,
        )
        pipeline_run2.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        create_pipeline_run_with_schedule(
            pipeline_uuid='test_active_run_1',
            pipeline_schedule_id=pipeline_schedule.id,
        )
        pipeline_run4 = create_pipeline_run_with_schedule(
            pipeline_uuid='test_active_run_2',
        )
        pipeline_run4.update(status=PipelineRun.PipelineRunStatus.RUNNING)
        results1 = PipelineRun.active_runs(
            pipeline_uuids=['test_active_run_1'],
            include_block_runs=True,
        )
        results2 = PipelineRun.active_runs(
            pipeline_uuids=['test_active_run_2'],
            include_block_runs=True,
        )
        results3 = PipelineRun.active_runs(
            include_block_runs=True,
        )
        self.assertEqual(
            set([r.id for r in results1]),
            set([pipeline_run.id, pipeline_run2.id]),
        )
        self.assertEqual(set([r.id for r in results2]), set([pipeline_run4.id]))
        self.assertEqual(
            set([r.id for r in results3]),
            set([pipeline_run.id, pipeline_run2.id, pipeline_run4.id]),
        )


class BlockRunTests(DBTestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()
        self.pipeline = create_pipeline_with_blocks(
            'test pipeline',
            self.repo_path,
        )

    def test_log_file(self):
        execution_date = datetime.now()
        pipeline_run = create_pipeline_run_with_schedule(
            pipeline_uuid='test_pipeline',
            execution_date=execution_date,
        )
        execution_date_str = execution_date.strftime(format='%Y%m%dT%H%M%S')
        for b in pipeline_run.block_runs:
            expected_file_path = os.path.join(
                get_repo_config(self.repo_path).variables_dir,
                'pipelines/test_pipeline/.logs',
                f'{pipeline_run.pipeline_schedule_id}/{execution_date_str}/{b.block_uuid}.log',
            )
            self.assertEqual(b.logs.get('path'), expected_file_path)



from mage_ai.api.presenters.BasePresenter import BasePresenter


class BlockRunPresenter(BasePresenter):
    default_attributes = [
        'block_uuid',
        'completed_at',
        'created_at',
        'id',
        'metrics',
        'pipeline_run_id',
        'pipeline_schedule_id',
        'pipeline_schedule_name',
        'started_at',
        'status',
        'updated_at',
    ]

    def present(self, **kwargs):
        return self.model.to_dict()



from enum import Enum

DBT_CLOUD_BASE_URL = 'https://cloud.getdbt.com/api/v2/accounts'

DEFAULT_POLL_INTERVAL = 3


class DbtCloudJobRunStatus(Enum):
    QUEUED = 1
    STARTING = 2
    RUNNING = 3
    SUCCESS = 10
    ERROR = 20
    CANCELLED = 30
    TERMINAL_STATUSES = (SUCCESS, ERROR, CANCELLED)



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.REMOVE

    Docs: https://docs.mage.ai/guides/transformer-blocks#remove-columns
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.REMOVE,
        arguments=[],  # Specify columns to remove
        axis=Axis.COLUMN,
    )
{% endblock %}



import traceback
from typing import Dict

from mage_ai.data_preparation.executors.pipeline_executor import PipelineExecutor
from mage_ai.services.k8s.job_manager import JobManager as K8sJobManager


class K8sPipelineExecutor(PipelineExecutor):
    def cancel(
        self,
        pipeline_run_id: int = None,
    ):
        if pipeline_run_id is None:
            return
        try:
            job_manager = K8sJobManager(
                job_name=f'mage-data-prep-pipeline-{pipeline_run_id}',
            )
            job_manager.delete_job()
        except Exception:
            traceback.print_exc()

    def execute(
        self,
        pipeline_run_id: int = None,
        global_vars: Dict = None,
        **kwargs,
    ) -> None:
        cmd = f'/app/run_app.sh '\
              f'python mage_ai/cli/main.py run {self.pipeline.repo_config.repo_path} '\
              f'{self.pipeline.uuid}'
        options = [
            '--executor-type local_python',
        ]
        if self.execution_partition is not None:
            options.append(f'--execution-partition {self.execution_partition}')
        if pipeline_run_id is not None:
            options.append(f'--pipeline-run-id {pipeline_run_id}')
        options_str = ' '.join(options)
        job_manager = K8sJobManager(
            job_name=f'mage-data-prep-pipeline-{pipeline_run_id}',
            logger=self.logger,
            logging_tags=kwargs.get('tags', dict()),
        )
        job_manager.run_job(f'{cmd} {options_str}')



from inspect import isawaitable
from mage_ai.api.presenters.BasePresenter import BasePresenter


class KernelPresenter(BasePresenter):
    default_attributes = [
        'alive',
        'id',
        'name',
        'usage',
    ]

    async def present(self, **kwargs):
        kernel = self.resource.model

        kernel_response = dict(
            alive=kernel.is_alive(),
            id=kernel.kernel_id,
            name=kernel.kernel_name,
        )
        if kernel.has_kernel:
            try:
                client = kernel.client()
                session = kernel.session
                control_channel = client.control_channel
                usage_request = session.msg("usage_request", {})

                control_channel.send(usage_request)
                res = client.control_channel.get_msg(timeout=3)
                if isawaitable(res):
                    # control_channel.get_msg may return a Future,
                    # depending on configured KernelManager class
                    res = await res
                kernel_response['usage'] = res.get('content')
            except Exception:
                pass

        return kernel_response



from datetime import datetime
from mage_ai.utils.logger.constants import (
    LOG_LEVEL_DEBUG,
    LOG_LEVEL_ERROR,
    LOG_LEVEL_EXCEPTION,
    LOG_LEVEL_INFO,
    TYPE_LOG,
)
from mage_ai.shared.parsers import encode_complex
from typing import Any
import simplejson
import sys
import uuid


class Logger():
    def __init__(
        self,
        caller: Any = None,
        log_to_stdout: bool = False,
        logger: Any = None,
        verbose: int = 1,
    ):
        self.caller = caller
        self.log_to_stdout = log_to_stdout
        self.logger = logger
        self.verbose = verbose

    def build_tags(self, **kwargs):
        return {k: v for k, v in kwargs.items() if v}

    def debug(self, message, tags={}, **kwargs):
        self.__log(LOG_LEVEL_DEBUG, message, tags, **kwargs)

    def error(self, message, tags={}, **kwargs):
        self.__log(LOG_LEVEL_ERROR, message, tags, **kwargs)

    def exception(self, message, tags={}, **kwargs):
        self.__log(LOG_LEVEL_EXCEPTION, message, tags, **kwargs)

    def info(self, message, tags={}, **kwargs):
        self.__log(LOG_LEVEL_INFO, message, tags, **kwargs)

    def __log(self, level, message, tags, **kwargs) -> None:
        if self.verbose == 0:
            return

        now = datetime.utcnow()
        if self.caller:
            if type(self.caller) is str:
                caller_string = self.caller
            else:
                caller_string = self.caller.__class__.__name__
        else:
            caller_string = self.__class__.__name__

        data = dict(
            caller=caller_string,
            level=level,
            message=message,
            tags=tags,
            timestamp=int(now.timestamp()),
            uuid=uuid.uuid4().hex,
        )
        data.update(kwargs)
        data.update(type=TYPE_LOG)

        json_string = simplejson.dumps(
            data,
            default=encode_complex,
            ignore_nan=True,
        )

        if self.log_to_stdout:
            sys.stdout.write(f'{json_string}\n')
            sys.stdout.flush()
        elif self.logger:
            # log
            # warn
            method = None
            if LOG_LEVEL_ERROR == level:
                method = 'error'
            elif LOG_LEVEL_EXCEPTION == level:
                method = 'exception'
            elif LOG_LEVEL_INFO == level:
                method = 'info'

            getattr(self.logger, method)(f'[{now.isoformat()}] {json_string}')
        else:
            print(json_string)



from mage_ai.api.middleware import parse_cookie_header
from mage_ai.api.utils import (
    has_at_least_editor_role,
    has_at_least_viewer_role,
    is_owner,
)
from mage_ai.authentication.passwords import create_bcrypt_hash, generate_salt
from mage_ai.orchestration.db.models.oauth import Permission, Role, User
from mage_ai.tests.base_test import DBTestCase


class UtilsTest(DBTestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()
        Role.create_default_roles()

    def test_cookie_parser(self):
        named_cookies = "csrftoken=5lkzK7FCI2iWy2xi7wbZPI7P26qbspIE; django_language=en"

        self.assertEqual(
            parse_cookie_header(named_cookies),
            {
                "csrftoken": "5lkzK7FCI2iWy2xi7wbZPI7P26qbspIE",
                "django_language": "en",
            },
        )

        unnamed_cookies = (
            "csrftoken=5lkzK7FCI2iWy2xi7wbZPI7P26qbspIE; unnamed; django_language=en"
        )

        self.assertEqual(
            parse_cookie_header(unnamed_cookies),
            {
                "csrftoken": "5lkzK7FCI2iWy2xi7wbZPI7P26qbspIE",
                "django_language": "en",
                "": "unnamed",
            },
        )

    def test_get_user_access_for_global_owner(self):
        password_salt = generate_salt()
        user = User.create(
            email='admin@admin.com',
            password_hash=create_bcrypt_hash('admin', password_salt),
            password_salt=password_salt,
            roles_new=[Role.query.filter(Role.name == 'Owner').first()],
            username='admin',
        )
        access = user.get_access(None)
        self.assertEqual(1, access)

        access = user.get_access(Permission.Entity.GLOBAL, None)
        self.assertEqual(1, access)

        access = user.get_access(Permission.Entity.PIPELINE, 'test')
        self.assertEqual(1, access)

        self.assertTrue(is_owner(user))

    def test_get_user_access_for_pipeline_editor(self):
        password_salt = generate_salt()
        user = User.create(
            email='editor@edit.com',
            password_hash=create_bcrypt_hash('edit', password_salt),
            password_salt=password_salt,
            roles_new=[Role.create(
                name='pipeline_editor',
                permissions=[
                    Permission.create(
                        entity=Permission.Entity.PIPELINE,
                        entity_id='test',
                        access=4,
                    )
                ]
            )],
            username='editor',
        )
        access = user.get_access(None)
        self.assertEqual(4, access)

        access = user.get_access(Permission.Entity.GLOBAL, None)
        self.assertEqual(0, access)

        access = user.get_access(Permission.Entity.PIPELINE, 'test')
        self.assertEqual(4, access)

        access = user.get_access(Permission.Entity.PIPELINE, 'not_test')
        self.assertEqual(0, access)

        self.assertTrue(has_at_least_editor_role(user, Permission.Entity.PIPELINE, 'test'))

    def test_get_user_access_for_role_with_multiple_permissions(self):
        password_salt = generate_salt()
        user = User.create(
            email='editor2@edit.com',
            password_hash=create_bcrypt_hash('edit', password_salt),
            password_salt=password_salt,
            roles_new=[Role.create(
                name='pipeline_editor',
                permissions=[
                    Permission.create(
                        entity=Permission.Entity.PIPELINE,
                        entity_id='test',
                        access=4,
                    )
                ]
            ), Role.create(
                name='pipeline_viewer',
                permissions=[
                    Permission.create(
                        entity=Permission.Entity.PIPELINE,
                        entity_id='test',
                        access=8,
                    ),
                    Permission.create(
                        entity=Permission.Entity.PIPELINE,
                        entity_id='test1',
                        access=8,
                    )
                ]
            )],
            username='editor2',
        )
        access = user.get_access(None)
        self.assertEqual(12, access)

        access = user.get_access(Permission.Entity.GLOBAL, None)
        self.assertEqual(0, access)

        access = user.get_access(Permission.Entity.PIPELINE, 'test')
        self.assertEqual(12, access)

        access = user.get_access(Permission.Entity.PIPELINE, 'test1')
        self.assertEqual(8, access)

        access = user.get_access(Permission.Entity.PIPELINE, 'not_test')
        self.assertEqual(0, access)

        self.assertTrue(has_at_least_editor_role(user, Permission.Entity.PIPELINE, 'test'))
        self.assertFalse(has_at_least_editor_role(user, Permission.Entity.PIPELINE, 'test1'))
        self.assertTrue(has_at_least_viewer_role(user, Permission.Entity.PIPELINE, 'test1'))



import os

import inflection

from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.repo_manager import MAGE_DATA_DIR_ENV_VAR
from mage_ai.data_preparation.shared.secrets import get_secret_value
from mage_ai.tests.api.operations.test_base import BaseApiTestCase
from mage_ai.tests.factory import create_user


class SyncOperationTests(BaseApiTestCase):
    @property
    def model_class_name(self) -> str:
        return 'sync'

    @property
    def model_class_name_plural(self) -> str:
        return inflection.pluralize(self.model_class_name)

    @classmethod
    def setUpClass(self):
        super().setUpClass()
        os.environ[MAGE_DATA_DIR_ENV_VAR] = self.repo_path

    async def asyncTearDown(self):
        # super().asyncTearDown()
        os.remove(get_preferences().preferences_file_path)

    async def test_execute_create_user(self):
        email = self.faker.email()
        user = create_user(email=email, roles=1)
        operation = self.build_create_operation(
            dict(
                remote_repo_link='test_link',
                repo_path=self.repo_path,
                branch='main',
                auth_type='https',
                user_git_settings=dict(
                    username='username1',
                    email=email,
                    access_token='abc123',
                ),
            ),
            user=user
        )
        response = await operation.execute()
        self.assertEqual(response['sync']['remote_repo_link'], 'test_link')

        preferences = get_preferences(user=user).sync_config
        self.assertEqual(preferences['remote_repo_link'], 'test_link')
        self.assertEqual(preferences['username'], 'username1')

        preferences = get_preferences().sync_config
        print('response:', preferences)
        self.assertEqual(preferences['remote_repo_link'], 'test_link')
        self.assertIsNone(preferences.get('username'))

        self.assertEqual(user.git_settings['username'], 'username1')
        self.assertEqual(
            get_secret_value(user.git_settings['access_token_secret_name']),
            'abc123',
        )

    async def test_execute_create_both(self):
        email = self.faker.email()
        user = create_user(email=email, roles=1)
        operation = self.build_create_operation(
            dict(
                remote_repo_link='test_link',
                repo_path=self.repo_path,
                branch='main',
                auth_type='https',
                username='username',
                email='admin@admin.com',
                access_token='abc123',
                user_git_settings=dict(
                    username='another_username',
                    email=email,
                    access_token='def456',
                )
            ),
            user=user,
        )
        response = await operation.execute()
        self.assertEqual(response['sync']['remote_repo_link'], 'test_link')

        preferences = get_preferences(user=user).sync_config
        project_preferences = get_preferences().sync_config
        self.assertEqual(preferences['username'], 'another_username')
        self.assertEqual(project_preferences['username'], 'username')
        self.assertEqual(user.git_settings['email'], email)
        self.assertEqual(
            get_secret_value(project_preferences['access_token_secret_name']),
            'abc123',
        )
        self.assertEqual(get_secret_value(user.git_settings['access_token_secret_name']), 'def456')



from mage_ai.data_cleaner.estimators.base import BaseEstimator
from sklearn.decomposition import PCA
from sklearn.ensemble import IsolationForest
from sklearn.neighbors import LocalOutlierFactor
import numpy as np

MAX_CONTAMINATION_RATE = 0.15
MIN_CONTAMINATION_RATE = 0.025
SUPPORTED_METHODS = ['lof', 'auto', 'itree']


class OutlierRemover(BaseEstimator):
    """
    Automatically removes outliers from the given data.
    """

    def __init__(self, method: str = 'auto') -> None:
        """
        Constructs an outlier remover.

        Args:
            method (str, optional): Specifies the outlier removal method to use. Defaults to
            'auto'. There are three options:
            - 'lof' - Local Outlier Factor
            - 'itree' - Isolation Forest
            - 'auto' - Automatically best method based on runtime comparisons
        """
        if method not in SUPPORTED_METHODS:
            raise ValueError(f'The method specified \'{method}\' is not supported.')
        self.method = method

    def estimate_contamination_rate(self, X: np.ndarray) -> float:
        """
        Guesses the outlier contamination rate from the data. Uses the IQR
        rule to estimate outliers dimensionwise, and then aggregates the entire
        count to provide an estimate on the number of outliers in the dataset. This
        estimate is used to compute the contamination factor used in finding outliers with
        more complex algorithms.

        The exact algorithm used to estimate the contamination rate is as follows:
        1. Calculate the first and third quantile for each dimension
        2. Calculate the interquartile range for each dimension
        3. Compute the per-dimension outlier region as x <= lower - 1.5*IQR and x >= upper + 1.5*IQR
        4. Find all entries that exist in the outlier region per-dimension
        5. Mark examples with at least 50% of its features as outliers to be potential
        multidimensional outliers
        6. Return the ratio of these potential outliers in the region

        Args:
            X (np.ndarray): Input array of shape (n_samples, n_dimensions)

        Returns:
            float: The guessed contamination rate
        """
        # TODO - use PCA to calculate the iqrs along the axes of greatest variance
        first = np.quantile(X, 0.25, axis=0)
        third = np.quantile(X, 0.75, axis=0)
        iqr_diffs = 1.5 * (third - first)

        lower_bound = first - iqr_diffs
        upper_bound = third + iqr_diffs

        # select all rows with at least 50% potential outliers
        mask = (X < lower_bound) | (X > upper_bound)
        mask = mask.sum(axis=1) / mask.shape[1]
        mask = mask >= 0.5
        return mask.sum() / mask.size

    def fit(self, X: np.ndarray, y: np.ndarray = None) -> None:
        """
        Fits the outlier remover on the given data.

        Args:
            X (np.ndarray): Input array of shape (n_samples, n_dimensions)
            y (np.ndarray, optional): Not used. Defaults to None.
        """
        count, ndim = X.shape
        pca_transformer = PCA(n_components=20, random_state=42)
        # Clamp contamination rate in case that IQR method gives bad results
        contamination_rate = max(self.estimate_contamination_rate(X), MIN_CONTAMINATION_RATE)
        contamination_rate = min(contamination_rate, MAX_CONTAMINATION_RATE)

        if ndim > 20:
            X = pca_transformer.fit_transform(X)
        if self.method == 'auto':
            if ndim <= 5:
                self.method = 'lof'
            else:
                self.method = 'itree'
        if self.method == 'lof':
            if count < 10:
                n_neighbors = 2
            elif count < 500:
                n_neighbors = count // 10 + 1
            elif contamination_rate <= 0.1:
                n_neighbors = 20
            else:
                n_neighbors = 35
            self.outlier_remover = LocalOutlierFactor(
                n_neighbors=n_neighbors,
                n_jobs=-1,
                contamination=contamination_rate,
            )
        elif self.method == 'itree':
            self.outlier_remover = IsolationForest(
                contamination=contamination_rate,
                n_estimators=100,
                n_jobs=-1,
                random_state=42,
            )
            self.outlier_remover.fit(X, y)

    def transform(self, X: np.ndarray, **kwargs) -> np.ndarray:
        """
        Determines the labels of outliers in the input data

        Args:
            X (np.ndarray): Input array of shape (n_samples, n_dimensions)

        Returns:
            np.ndarray: Array of shape (n_samples,) containing labels for the entries.
            Label is True if outlier, else False.
        """
        if self.method == 'lof':
            labels = self.outlier_remover.fit_predict(X, **kwargs)
        else:
            labels = self.outlier_remover.predict(X, **kwargs)
        return labels == -1



from os import path

from pandas import DataFrame

from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.mongodb import MongoDB

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_mongodb(df: DataFrame, **kwargs) -> None:
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    MongoDB.with_config(ConfigFileLoader(config_path, config_profile)).export(
        df,
        collection='collection_name',
    )



from datetime import datetime
from mage_ai.api.errors import ApiError
from mage_ai.api.resources.BaseResource import BaseResource
from mage_ai.authentication.oauth2 import encode_token, generate_access_token
from mage_ai.authentication.passwords import verify_password
from mage_ai.authentication.ldap import new_ldap_connection
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.settings import AUTHENTICATION_MODE
from mage_ai.usage_statistics.logger import UsageStatisticLogger


class SessionResource(BaseResource):
    @classmethod
    @safe_db_query
    async def create(self, payload, _, **kwargs):
        email = payload.get('email')
        password = payload.get('password')
        username = payload.get('username')

        error = ApiError.RESOURCE_NOT_FOUND
        error.update({'message': 'Email/username and/or password invalid.'})

        if not (email or username) or not password:
            error.update(
                {'message': 'Email/username and password are required.'})
            raise ApiError(error)

        async def _create_callback(resource):
            await UsageStatisticLogger().users_impression()

        self.on_create_callback = _create_callback

        user = None
        if AUTHENTICATION_MODE.lower() == 'ldap':
            # we can use just the method verify here authz=verify(username,password)
            conn = new_ldap_connection()
            auth, user_dn = conn.authenticate(email, password)
            if not auth:
                if user_dn != "":
                    error.update({'message': 'wrong password.'})
                raise ApiError(error)

            authz = conn.authorize(user_dn)
            if not authz:
                error.update(
                        {'message': 'user not authorized. contact your admin'})
                raise ApiError(error)
            if email:
                user = User.query.filter(User.username == email).first()
            if not user:  # noqa: E712
                print('first user login, creating user.')
                user = User.create(
                    roles=2,
                    username=email,
                )
            oauth_token = generate_access_token(user, kwargs['oauth_client'])
            return self(oauth_token, user, **kwargs)

        if email:
            user = User.query.filter(User.email == email).first()
        elif username:
            user = User.query.filter(User.username == username).first()
        if not user:
            raise ApiError(error)

        if verify_password(password, user.password_hash):
            oauth_token = generate_access_token(user, kwargs['oauth_client'])
            return self(oauth_token, user, **kwargs)
        else:
            raise ApiError(error)

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        return self(kwargs['oauth_token'], user, **kwargs)

    @safe_db_query
    def update(self, payload, **kwargs):
        self.model.expires = datetime.utcnow()
        self.model.save()

    @safe_db_query
    def token(self):
        return encode_token(self.model.token, self.model.expires)



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Dict, List
import json
import os
import traceback


class BaseSink(ABC):
    config_class = None

    def __init__(self, config: Dict, **kwargs):
        if self.config_class is not None:
            if 'connector_type' in config:
                config.pop('connector_type')
            self.config = self.config_class.load(config=config)
        self.buffer_path = kwargs.get('buffer_path')
        self.buffer = self.read_buffer() or []
        self.buffer_start_time = None
        self.init_client()

    def init_client(self):
        pass

    @abstractmethod
    def write(self, data: Dict):
        pass

    @abstractmethod
    def batch_write(self, data: List[Dict]):
        pass

    def clear_buffer(self):
        self.buffer = []
        if not self.buffer_path:
            return
        with open(self.buffer_path, 'w'):
            pass

    def has_buffer_timed_out(self, buffer_timeout_seconds):
        if self.buffer_start_time is None:
            return False
        if buffer_timeout_seconds <= 0:
            return False
        return (datetime.now(timezone.utc) -
                self.buffer_start_time).total_seconds() >= buffer_timeout_seconds

    def read_buffer(self):
        buffer = []
        if not self.buffer_path or not os.path.exists(self.buffer_path):
            return buffer
        try:
            with open(self.buffer_path) as fp:
                for line in fp:
                    buffer.append(json.loads(line))
        except Exception:
            traceback.print_exc()
            pass
        return buffer

    def test_connection(self):
        return True

    def write_buffer(self, data: List[Dict]):
        if not data:
            return
        if not self.buffer:
            self.buffer_start_time = datetime.now(timezone.utc)
        self.buffer += data
        if not self.buffer_path or not data:
            return
        with open(self.buffer_path, 'a') as fp:
            for record in data:
                fp.write(json.dumps(record) + '\n')

    def _print(self, msg):
        print(f'[{self.__class__.__name__}] {msg}')



from azure.eventhub import EventHubConsumerClient
from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sources.base import BaseSource
from typing import Callable, List
import traceback


@dataclass
class AzureEventHubConfig(BaseConfig):
    connection_str: str
    eventhub_name: str
    consumer_group: str = '$Default'


class AzureEventHubSource(BaseSource):
    config_class = AzureEventHubConfig

    def init_client(self):
        self.consumer_client = EventHubConsumerClient.from_connection_string(
            conn_str=self.config.connection_str,
            consumer_group=self.config.consumer_group,
            eventhub_name=self.config.eventhub_name,
        )

    def read(self, handler: Callable):
        try:
            def on_event(partition_context, event):
                self._print(f'Received event from partition: {partition_context.partition_id}.')
                self._print(f'Event: {event}')

                handler(dict(data=event.body_as_str()))

            with self.consumer_client:
                self.consumer_client.receive(
                    on_event=on_event,
                    on_partition_initialize=self.on_partition_initialize,
                    on_partition_close=self.on_partition_close,
                    on_error=self.on_error,
                    starting_position='-1',  # '-1' is from the beginning of the partition.
                )
        except KeyboardInterrupt:
            self._print('Stopped receiving.')

    def batch_read(self, handler: Callable):
        try:
            def on_event_batch(partition_context, event_batch: List):
                if len(event_batch) == 0:
                    return
                self._print(f'Partition {partition_context.partition_id},'
                            f'Received count: {len(event_batch)}')
                self._print(f'Sample event: {event_batch[0]}')

                # Handle events
                try:
                    handler([dict(data=e.body_as_str()) for e in event_batch])
                except Exception as e:
                    traceback.print_exc()
                    raise e

                partition_context.update_checkpoint()

            with self.consumer_client:
                self.consumer_client.receive_batch(
                    on_event_batch=on_event_batch,
                    max_batch_size=100,
                    on_partition_initialize=self.on_partition_initialize,
                    on_partition_close=self.on_partition_close,
                    on_error=self.on_error,
                    starting_position='-1',  # '-1' is from the beginning of the partition.
                )
        except KeyboardInterrupt:
            self._print('Stopped receiving.')

    def test_connection(self):
        return True

    def on_partition_initialize(self, partition_context):
        self._print(f'Partition: {partition_context.partition_id} has been initialized.')

    def on_partition_close(self, partition_context, reason):
        self._print(f'Partition: {partition_context.partition_id} has been closed, '
                    f'reason for closing: {reason}.')

    def on_error(self, partition_context, error):
        if partition_context:
            self._print(f'An exception: {partition_context.partition_id} occurred during'
                        f' receiving from Partition: {error}.')
        else:
            self._print(f'An exception: {error} occurred during the load balance process.')



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.CLEAN_COLUMN_NAME

    Docs: https://docs.mage.ai/guides/transformer-blocks#clean-column-names
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.CLEAN_COLUMN_NAME,
        arguments=df.columns,
        axis=Axis.COLUMN,
    )
{% endblock %}



"""Create users, oauth2 applications, and oauth2 access tokens

Revision ID: 97ff9f55a3c0
Revises: 1d8e65aeefdd
Create Date: 2023-01-31 21:19:33.645918

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '97ff9f55a3c0'
down_revision = '1d8e65aeefdd'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('user',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('avatar', sa.String(length=255), nullable=True),
    sa.Column('email', sa.String(length=255), nullable=True),
    sa.Column('first_name', sa.String(length=255), nullable=True),
    sa.Column('last_name', sa.String(length=255), nullable=True),
    sa.Column('owner', sa.Boolean(), nullable=True),
    sa.Column('password_hash', sa.String(length=255), nullable=True),
    sa.Column('password_salt', sa.String(length=255), nullable=True),
    sa.Column('roles', sa.Integer(), nullable=True),
    sa.Column('username', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('oauth2_application',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('authorization_grant_type', sa.Enum('AUTHORIZATION_CODE', 'CLIENT_CREDENTIALS', name='authorizationgranttype'), nullable=True),
    sa.Column('client_id', sa.String(length=255), nullable=True),
    sa.Column('client_type', sa.Enum('PRIVATE', 'PUBLIC', name='clienttype'), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('redirect_uris', sa.String(length=255), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('oauth2_access_token',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('expires', sa.DateTime(timezone=True), nullable=True),
    sa.Column('oauth2_application_id', sa.Integer(), nullable=True),
    sa.Column('token', sa.String(length=255), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['oauth2_application_id'], ['oauth2_application.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('oauth2_access_token')
    op.drop_table('oauth2_application')
    op.drop_table('user')
    # ### end Alembic commands ###



from typing import Any, Dict, List, Union
import json


def parse_runtime_variables(variables: List[str]) -> Dict[str, Any]:
    """
    Returns a dictionary of variable to parsed values.
    """
    vars_parsed = dict()

    if variables is None:
        return vars_parsed
    for i in range(0, len(variables), 2):
        key = variables[i]
        try:
            value = variables[i + 1]
        except IndexError:
            value = None
        vars_parsed[key] = get_value(value)
    return vars_parsed


def get_value(value: Union[str, None]) -> Any:
    """
    Returns parsed value.

    If the string is a valid JSON object, deserializes it to a Python object.
    Otherwise, passes the original value.
    """
    if value is not None:
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            pass
    return value



import asyncio
import base64
import os
import shutil
import subprocess
import uuid
from typing import Any, List
from urllib.parse import urlparse, urlsplit, urlunsplit

from mage_ai.data_preparation.preferences import get_preferences
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.data_preparation.shared.secrets import get_secret_value
from mage_ai.data_preparation.sync import AuthType, GitConfig
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.shared.logger import VerboseFunctionExec

DEFAULT_SSH_KEY_DIRECTORY = os.path.expanduser('~/.ssh')
REMOTE_NAME = 'mage-repo'

# Git authentication variables
GIT_SSH_PUBLIC_KEY_VAR = 'GIT_SSH_PUBLIC_KEY'
GIT_SSH_PRIVATE_KEY_VAR = 'GIT_SSH_PRIVATE_KEY'
GIT_ACCESS_TOKEN_VAR = 'GIT_ACCESS_TOKEN'


class Git:
    def __init__(self, git_config: GitConfig) -> None:
        import git
        self.remote_repo_link = git_config.remote_repo_link
        self.repo_path = git_config.repo_path or os.getcwd()
        os.makedirs(self.repo_path, exist_ok=True)
        self.git_config = git_config
        self.auth_type = git_config.auth_type or AuthType.SSH

        if self.auth_type == AuthType.HTTPS:
            url = urlsplit(self.remote_repo_link)
            if os.getenv(GIT_ACCESS_TOKEN_VAR):
                token = os.getenv(GIT_ACCESS_TOKEN_VAR)
            else:
                token = get_secret_value(
                    git_config.access_token_secret_name,
                    repo_name=get_repo_path(),
                )
            user = git_config.username
            url = url._replace(netloc=f'{user}:{token}@{url.netloc}')
            self.remote_repo_link = urlunsplit(url)

        try:
            self.repo = git.Repo(self.repo_path)
        except git.exc.InvalidGitRepositoryError:
            self.__setup_repo()

        self.__set_git_config()

        try:
            self.repo.create_remote(REMOTE_NAME, self.remote_repo_link)
        except git.exc.GitCommandError:
            # if the remote already exists
            pass

        # replace the existing remote url if it is different from the provided url
        self.origin = self.repo.remotes[REMOTE_NAME]
        if self.remote_repo_link not in self.origin.urls:
            self.origin.set_url(self.remote_repo_link)

    @classmethod
    def get_manager(self, user: User = None) -> 'Git':
        preferences = get_preferences(user=user)
        git_config = GitConfig.load(config=preferences.sync_config)
        return Git(git_config)

    @property
    def current_branch(self) -> Any:
        return self.repo.git.branch('--show-current')

    @property
    def branches(self) -> List:
        return [branch.name for branch in self.repo.branches]

    def untracked_files(self, untracked_files: bool = False) -> List[str]:
        from git.compat import defenc

        # ---------- Taken from GitPython source code -----------
        proc = self.repo.git.status(
            as_process=True,
            porcelain=True,
            untracked_files=untracked_files,
        )
        # Untracked files prefix in porcelain mode
        prefix = '?? '
        untracked_files = []
        for line in proc.stdout:
            line = line.decode(defenc)
            if not line.startswith(prefix):
                continue
            filename = line[len(prefix):].rstrip('\n')
            # Special characters are escaped
            if filename[0] == filename[-1] == '"':
                filename = filename[1:-1]
                # WHATEVER ... it's a mess, but works for me
                filename = (
                    filename
                    .encode('ascii')
                    .decode('unicode_escape')
                    .encode('latin1')
                    .decode(defenc)
                )
            untracked_files.append(filename)
        proc.wait()
        # -------------------------------------------------------
        return untracked_files

    @property
    def modified_files(self) -> List[str]:
        return [item.a_path for item in self.repo.index.diff(None)]

    async def check_connection(self) -> None:
        proc = self.repo.git.ls_remote(self.origin.name, as_process=True)

        self.__poll_process_with_timeout(
            proc,
            error_message='Error connecting to remote, make sure your SSH key is set up properly.',
        )

    def _run_command(self, command: str) -> None:
        proc = subprocess.Popen(args=command, shell=True)
        proc.wait()

    def _remote_command(func) -> None:
        '''
        Decorator method for commands that need to connect to the remote repo. This decorator
        will configure and test SSH settings before executing the Git command.
        '''
        def wrapper(self, *args, **kwargs):
            def add_host_to_known_hosts():
                self.__add_host_to_known_hosts()
                asyncio.run(self.check_connection())

            if self.auth_type == AuthType.SSH:
                url = f'ssh://{self.git_config.remote_repo_link}'
                hostname = urlparse(url).hostname

                private_key_file = self.__create_ssh_keys()
                git_ssh_cmd = f'ssh -i {private_key_file}'
                with self.repo.git.custom_environment(GIT_SSH_COMMAND=git_ssh_cmd):
                    try:
                        asyncio.run(self.check_connection())
                    except ChildProcessError as err:
                        if 'Host key verification failed' in str(err):
                            if hostname:
                                add_host_to_known_hosts()
                        else:
                            raise err
                    except TimeoutError:
                        if hostname:
                            add_host_to_known_hosts()
                        else:
                            raise TimeoutError(
                                "Connecting to remote timed out, make sure your SSH key is set up properly"  # noqa: E501
                                " and your repository host is added as a known host. More information here:"  # noqa: E501
                                " https://docs.mage.ai/developing-in-the-cloud/setting-up-git#5-add-github-com-to-known-hosts")  # noqa: E501
                    func(self, *args, **kwargs)
            else:
                asyncio.run(self.check_connection())
                func(self, *args, **kwargs)

        return wrapper

    @_remote_command
    def reset(self, branch: str = None) -> None:
        self.origin.fetch()
        if branch is None:
            branch = self.current_branch
        self.repo.git.reset('--hard', f'{self.origin.name}/{branch}')
        self.__pip_install()

    @_remote_command
    def push(self) -> None:
        self.repo.git.push(
            '--set-upstream',
            self.origin.name,
            self.current_branch
        )

    @_remote_command
    def pull(self) -> None:
        self.origin.pull(self.current_branch)
        self.__pip_install()

    def status(self) -> str:
        return self.repo.git.status()

    def commit(self, message, files: List[str] = None) -> None:
        if self.repo.index.diff(None) or self.repo.untracked_files:
            if files:
                for file in files:
                    self.repo.git.add(file)
            else:
                self.repo.git.add('.')
            self.repo.index.commit(message)

    def switch_branch(self, branch) -> None:
        if branch in self.repo.heads:
            self.repo.git.switch(branch)
        else:
            self.repo.git.switch('-c', branch)

    @_remote_command
    def clone(self):
        from git import Repo
        tmp_path = f'{self.repo_path}_{str(uuid.uuid4())}'
        os.mkdir(tmp_path)
        try:
            # Clone remote repo to a tmp folder, and copy over the contents
            # of the tmp folder to the local repo.
            env = {}
            if self.auth_type == AuthType.SSH:
                private_key_file = self.__create_ssh_keys()
                env = {'GIT_SSH_COMMAND': f'ssh -i {private_key_file}'}
            Repo.clone_from(
                self.remote_repo_link,
                to_path=tmp_path,
                origin=REMOTE_NAME,
                env=env,
            )

            shutil.rmtree(os.path.join(self.repo_path, '.git'))
            shutil.copytree(
                tmp_path,
                self.repo_path,
                dirs_exist_ok=True,
                ignore=lambda x, y: ['.preferences.yaml']
            )
            self.repo.git.clean('-fd', exclude='.preferences.yaml')
            self.__pip_install()
        finally:
            shutil.rmtree(tmp_path)

    def __set_git_config(self):
        if self.git_config.username:
            self.repo.config_writer().set_value(
                'user', 'name', self.git_config.username).release()
        if self.git_config.email:
            self.repo.config_writer().set_value(
                'user', 'email', self.git_config.email).release()
        self.repo.config_writer('global').set_value(
            'safe', 'directory', self.repo_path).release()

    def __pip_install(self) -> None:
        requirements_file = os.path.join(
            self.repo.working_dir, 'requirements.txt')

        with VerboseFunctionExec(
            f'Running "pip3 install -r {requirements_file}"',
            verbose=True,
        ):
            try:
                if os.path.exists(requirements_file):
                    cmd = f'pip3 install -r {requirements_file}'
                    self._run_command(cmd)
                print(f'Installing {requirements_file} completed successfully.')
            except Exception as err:
                print(f'Skip installing {requirements_file} due to error: {err}')
                pass

    def __create_ssh_keys(self) -> str:
        if not os.path.exists(DEFAULT_SSH_KEY_DIRECTORY):
            os.mkdir(DEFAULT_SSH_KEY_DIRECTORY, 0o700)
        pubk_secret_name = self.git_config.ssh_public_key_secret_name
        if pubk_secret_name:
            public_key_file = os.path.join(
                DEFAULT_SSH_KEY_DIRECTORY,
                f'id_rsa_{pubk_secret_name}.pub'
            )
            if not os.path.exists(public_key_file):
                try:
                    public_key = get_secret_value(
                        pubk_secret_name,
                        repo_name=get_repo_path(),
                    )
                    if os.getenv(GIT_SSH_PUBLIC_KEY_VAR):
                        public_key = os.getenv(GIT_SSH_PUBLIC_KEY_VAR)
                    if public_key:
                        with open(public_key_file, 'w') as f:
                            f.write(base64.b64decode(public_key).decode('utf-8'))
                        os.chmod(public_key_file, 0o600)
                except Exception:
                    pass
        pk_secret_name = self.git_config.ssh_private_key_secret_name
        private_key_file = os.path.join(DEFAULT_SSH_KEY_DIRECTORY, 'id_rsa')
        if pk_secret_name:
            custom_private_key_file = os.path.join(
                DEFAULT_SSH_KEY_DIRECTORY,
                f'id_rsa_{pk_secret_name}'
            )
            if not os.path.exists(custom_private_key_file):
                try:
                    private_key = get_secret_value(
                        pk_secret_name,
                        repo_name=get_repo_path(),
                    )
                    if os.getenv(GIT_SSH_PRIVATE_KEY_VAR):
                        private_key = os.getenv(GIT_SSH_PRIVATE_KEY_VAR)
                    if private_key:
                        with open(custom_private_key_file, 'w') as f:
                            f.write(base64.b64decode(private_key).decode('utf-8'))
                        os.chmod(custom_private_key_file, 0o600)
                        private_key_file = custom_private_key_file
                except Exception:
                    pass
            else:
                private_key_file = custom_private_key_file

        return private_key_file

    def __setup_repo(self):
        import git
        tmp_path = f'{self.repo_path}_{str(uuid.uuid4())}'
        os.mkdir(tmp_path)
        try:
            # Clone the remote repo and copy over the .git folder
            # to initialize the local repository.
            env = {}
            if self.auth_type == AuthType.SSH:
                private_key_file = self.__create_ssh_keys()
                env = {'GIT_SSH_COMMAND': f'ssh -i {private_key_file}'}
                if not self.__add_host_to_known_hosts():
                    raise Exception('Could not add host to known_hosts')
            repo_git = git.cmd.Git(self.repo_path)
            repo_git.update_environment(**env)
            proc = repo_git.clone(
                self.remote_repo_link,
                tmp_path,
                origin=REMOTE_NAME,
                as_process=True,
            )

            asyncio.run(
                self.__poll_process_with_timeout(
                    proc,
                    error_message='Error cloning repo.',
                    timeout=20,
                )
            )

            git_folder = os.path.join(self.repo_path, '.git')
            tmp_git_folder = os.path.join(tmp_path, '.git')
            shutil.move(tmp_git_folder, git_folder)
            self.repo = git.Repo(path=self.repo_path)
        except Exception:
            self.repo = git.Repo.init(self.repo_path)
            # need to commit something to initialize the repository
            self.commit('Initial commit')
        finally:
            shutil.rmtree(tmp_path)

    def __add_host_to_known_hosts(self):
        url = f'ssh://{self.git_config.remote_repo_link}'
        hostname = urlparse(url).hostname
        if hostname:
            cmd = f'ssh-keyscan -t rsa {hostname} >> ~/.ssh/known_hosts'
            self._run_command(cmd)
            return True
        return False

    async def __poll_process_with_timeout(
        self,
        proc: subprocess.Popen,
        error_message: str = None,
        timeout: int = 10,
    ):
        ct = 0
        while ct < timeout * 2:
            return_code = proc.poll()
            if return_code is not None:
                proc.kill()
                break
            ct += 1
            await asyncio.sleep(0.5)

        if error_message is None:
            error_message = 'Error running Git process'

        if return_code is not None and return_code != 0:
            _, err = proc.communicate()
            message = (
                err.decode('UTF-8') if err
                else error_message
            )
            raise ChildProcessError(message)

        if return_code is None:
            proc.kill()
            raise TimeoutError



{% extends 'data_loaders/deltalake/default.jinja' %}
{% block storage_options %}
    storage_options = {
        'GOOGLE_SERVICE_ACCOUNT': '',
        'GOOGLE_SERVICE_ACCOUNT_PATH': '',
        'GOOGLE_SERVICE_ACCOUNT_KEY': '',
        'GOOGLE_BUCKET': '',
    }
{% endblock %}

{% block uri %}
    uri = 'gs://[bucket]/[key]'
{% endblock %}



from mage_ai.api.errors import ApiError
from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db import safe_db_query
from mage_ai.server.active_kernel import get_active_kernel_name
from mage_ai.server.api.clusters import ClusterType
from mage_ai.server.kernels import KernelName


class ClusterResource(GenericResource):
    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        clusters = []
        if ClusterType.EMR == pk and get_active_kernel_name() == KernelName.PYSPARK:
            from mage_ai.cluster_manager.aws.emr_cluster_manager import emr_cluster_manager

            clusters = emr_cluster_manager.list_clusters()

        return self(dict(
            clusters=clusters,
            type=pk,
        ), user, **kwargs)

    @safe_db_query
    def update(self, payload, **kwargs):
        if ClusterType.EMR == self.model['type']:
            from mage_ai.cluster_manager.aws.emr_cluster_manager import emr_cluster_manager

            error = ApiError.RESOURCE_INVALID.copy()

            if payload is None:
                error['message'] = 'Please include cluster info in the request payload.'
                raise ApiError(error)

            cluster_id = payload.get('id')
            if cluster_id is None:
                error['message'] = 'Please include cluster_id in the request payload.'
                raise ApiError(error)

            emr_cluster_manager.set_active_cluster(cluster_id)
            success = True

            self.model.update(
                id=cluster_id,
                success=success,
            )
            self.model.update(payload)



"""Create backfills table

Revision ID: 26305e46df52
Revises: 6aecc9bc451c
Create Date: 2023-01-27 10:50:04.283588

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '26305e46df52'
down_revision = '6aecc9bc451c'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('backfill',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('block_uuid', sa.String(length=255), nullable=True),
    sa.Column('completed_at', sa.DateTime(timezone=True), nullable=True),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('end_datetime', sa.DateTime(timezone=True), nullable=True),
    sa.Column('failed_at', sa.DateTime(timezone=True), nullable=True),
    sa.Column('interval_type', sa.Enum('SECOND', 'MINUTE', 'HOUR', 'DAY', 'WEEK', 'MONTH', 'YEAR', 'CUSTOM', name='intervaltype'), nullable=True),
    sa.Column('interval_units', sa.Integer(), nullable=True),
    sa.Column('metrics', sa.JSON(), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('pipeline_schedule_id', sa.Integer(), nullable=True),
    sa.Column('pipeline_uuid', sa.String(length=255), nullable=True),
    sa.Column('start_datetime', sa.DateTime(timezone=True), nullable=True),
    sa.Column('started_at', sa.DateTime(timezone=True), nullable=True),
    sa.Column('status', sa.Enum('INITIAL', 'RUNNING', 'COMPLETED', 'FAILED', 'CANCELLED', name='status'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('variables', sa.JSON(), nullable=True),
    sa.ForeignKeyConstraint(['pipeline_schedule_id'], ['pipeline_schedule.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    with op.batch_alter_table('pipeline_run', schema=None) as batch_op:
        batch_op.add_column(sa.Column('backfill_id', sa.Integer(), nullable=True))
        batch_op.create_foreign_key('pipeline_run_backfill_id', 'backfill', ['backfill_id'], ['id'])

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('pipeline_run', schema=None) as batch_op:
        batch_op.drop_constraint('pipeline_run_backfill_id', type_='foreignkey')
        batch_op.drop_column('backfill_id')

    op.drop_table('backfill')
    # ### end Alembic commands ###



from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.shared.hash import merge_dict
from mage_ai.tests.base_test import TestCase
from mage_ai.tests.data_cleaner.transformer_actions.shared import TEST_ACTION
import numpy as np
import pandas as pd


def build_df():
    return pd.DataFrame([
        [2, False, 5.0],
        ['$3', False, '$6.0', 1],
        ['$4,000', None, '$7,000', 200],
        ['$3', False, '$4.0', 3],
        ['$4,000', None, 3.0, 4],
        [5, True, 8000, 5],
    ], columns=['deposited', 'fund', 'amount', 'index']).set_index('index')


class BaseActionTests(TestCase):
    # def test_execute(self):
    #     df = build_df()

    #     base_action = BaseAction(merge_dict(
    #         TEST_ACTION,
    #         dict(action_code='%{1_1} >= 3 and (%{1_2} == false or %{1_2} != %{1_2}) '
    #                          'and %{1_4} >= 5.0'),
    #     ))

    #     self.assertEqual(base_action.execute(df).sort_values('deposited').to_numpy().tolist(), [
    #         ['$3', False, '$6.0'],
    #         ['$4,000', None, '$7,000'],
    #     ])

    def test_execute_axis_column(self):
        df = build_df()

        base_action = BaseAction(merge_dict(
            TEST_ACTION,
            dict(
                action_arguments=[
                    '%{1_1}',
                    # '%{3_1}',
                ],
                action_type='remove',
                axis='column',
            ),
        ))

        df_new = base_action.execute(df)
        self.assertEqual(df_new.values.tolist(), [
            [False, 5.0],
            [False, '$6.0'],
            [None, '$7,000'],
            [False, '$4.0'],
            [None, 3.0],
            [True, 8000],
        ])

    def test_execute_with_no_columns_to_transform(self):
        df = build_df()

        base_action = BaseAction(merge_dict(
            TEST_ACTION,
            dict(
                action_arguments=[
                    '%{1_1}',
                ],
                action_type='remove',
                axis='column',
            ),
        ))

        raised = False
        try:
            base_action.execute(df.drop(columns=['deposited']))
        except Exception:
            raised = True

        self.assertFalse(raised)

    def test_groupby(self):
        df = pd.DataFrame([
            ['a', '2020-01-03', 1050],
            ['a', '2020-01-01', 1000],
            ['b', '2020-01-04', 990],
            ['a', '2020-01-02', 1100],
            ['b', '2020-01-03', 1200],
        ], columns=[
            'store',
            'date',
            'sold',
        ])
        base_action = BaseAction(dict(
            action_type='group',
            action_arguments=['store'],
            action_code='',
            action_variables=dict(),
            child_actions=[
                dict(
                    action_type='sort',
                    axis='row',
                    action_arguments=['date'],
                    action_code='',
                    action_variables=dict(),
                ),
                dict(
                    action_type='diff',
                    action_arguments=['sold'],
                    action_code='',
                    action_variables=dict(),
                    axis='column',
                    outputs=[dict(uuid='sold_diff')]
                ),
                dict(
                    action_type='shift_down',
                    action_arguments=['sold'],
                    action_code='',
                    action_variables=dict(),
                    axis='column',
                    outputs=[dict(uuid='prev_sold')]
                ),
            ],
        ))
        df_new = base_action.execute(df)
        df_new = df_new.fillna(0)
        self.assertEqual(df_new.values.tolist(), [
            ['a', '2020-01-01', 1000, 0, 0],
            ['a', '2020-01-02', 1100, 100, 1000],
            ['a', '2020-01-03', 1050, -50, 1100],
            ['b', '2020-01-03', 1200, 0, 0],
            ['b', '2020-01-04', 990, -210, 1200],
        ])

    def test_hydrate_action(self):
        base_action = BaseAction(TEST_ACTION)
        base_action.hydrate_action()

        hydrated_action = TEST_ACTION.copy()
        hydrated_action['action_code'] = \
            'omni.deposited == True and (omni.fund == "The Quant" or omni.fund == "Yield")'
        hydrated_action['action_arguments'] = [
            'omni.deposited',
            'magic.spell',
        ]
        hydrated_action['action_options'] = dict(
            condition='omni.delivered_at >= magic.booked_at and magic.booked_at '
                      '>= omni.delivered_at - 2592000',
            default=0,
            timestamp_feature_a='omni.fund',
            timestamp_feature_b='omni.delivered_at',
            window=2592000,
        )

        self.assertEqual(base_action.action, hydrated_action)

    def test_hydrate_action_when_adding_column(self):
        base_action = BaseAction(merge_dict(TEST_ACTION, dict(
            action_type='add',
            axis='column',
        )))
        base_action.hydrate_action()

        hydrated_action = TEST_ACTION.copy()
        hydrated_action['action_code'] = \
            'omni.deposited == True and (omni.fund == "The Quant" or omni.fund == "Yield")'
        hydrated_action['action_type'] = 'add'
        hydrated_action['axis'] = 'column'
        hydrated_action['action_arguments'] = [
            'omni.deposited',
            'magic.spell',
        ]
        hydrated_action['action_options'] = dict(
            condition='omni.delivered_at >= magic.booked_at and magic.booked_at >= '
                      'omni.delivered_at - 2592000',
            default=0,
            timestamp_feature_a='omni.fund',
            timestamp_feature_b='omni.delivered_at',
            window=2592000,
        )

        self.assertEqual(base_action.action, hydrated_action)

    def test_join(self):
        df1 = pd.DataFrame([
            ['a', '2020-01-03', 1050],
            ['a', '2020-01-01', 1000],
            ['b', '2020-01-04', 990],
            ['a', '2020-01-02', 1100],
            ['b', '2020-01-03', 1200],
            ['c', '2020-01-07', 1250],
        ], columns=[
            'store',
            'date',
            'sold',
        ])
        df2 = pd.DataFrame([
            ['a', 'Store A'],
            ['b', 'Store B'],
        ], columns=[
            'store_name',
            'description',
        ])
        base_action = BaseAction(dict(
            action_type='join',
            action_arguments=[100],
            action_code='',
            action_options=dict(
                left_on=['store'],
                right_on=['store_name'],
                drop_columns=['store_name'],
                rename_columns={'description': 'store_description'}
            ),
            action_variables=dict(),
        ))
        df_new = base_action.execute(df1, df_to_join=df2)
        self.assertEqual(df_new.values.tolist(), [
            ['a', '2020-01-03', 1050, 'Store A'],
            ['a', '2020-01-01', 1000, 'Store A'],
            ['b', '2020-01-04', 990, 'Store B'],
            ['a', '2020-01-02', 1100, 'Store A'],
            ['b', '2020-01-03', 1200, 'Store B'],
            ['c', '2020-01-07', 1250, np.NaN],
        ])
        self.assertEqual(df_new.columns.to_list(), [
            'store', 'date', 'sold', 'store_description',
        ])

    def test_join_rename_column(self):
        df1 = pd.DataFrame([
            ['a', '2020-01-03', 1050],
            ['a', '2020-01-01', 1000],
            ['b', '2020-01-04', 990],
            ['a', '2020-01-02', 1100],
            ['b', '2020-01-03', 1200],
            ['c', '2020-01-07', 1250],
        ], columns=[
            'store',
            'date',
            'sold',
        ])
        df2 = pd.DataFrame([
            ['a', 'Store A', '2020-02-01'],
            ['b', 'Store B', '2020-02-02'],
        ], columns=[
            'store_name',
            'description',
            'date',
        ])
        base_action = BaseAction(dict(
            action_type='join',
            action_arguments=[100],
            action_code='',
            action_options=dict(
                left_on=['store'],
                right_on=['store_name'],
                drop_columns=['store_name'],
                rename_columns={'description': 'store_description'}
            ),
            action_variables=dict(),
            outputs=[
                {
                    'source_feature': {
                        'uuid': 'store_name',
                    },
                    'uuid': 'store_name',
                },
                {
                    'source_feature': {
                        'uuid': 'description',
                    },
                    'uuid': 'description',
                },
                {
                    'source_feature': {
                        'uuid': 'date',
                    },
                    'uuid': 'date_1',
                }
            ]
        ))
        df_new = base_action.execute(df1, df_to_join=df2)
        self.assertEqual(df_new.values.tolist(), [
            ['a', '2020-01-03', 1050, 'Store A', '2020-02-01'],
            ['a', '2020-01-01', 1000, 'Store A', '2020-02-01'],
            ['b', '2020-01-04', 990, 'Store B', '2020-02-02'],
            ['a', '2020-01-02', 1100, 'Store A', '2020-02-01'],
            ['b', '2020-01-03', 1200, 'Store B', '2020-02-02'],
            ['c', '2020-01-07', 1250, np.NaN, np.NaN],
        ])
        self.assertEqual(df_new.columns.to_list(), [
            'store', 'date', 'sold', 'store_description', 'date_1',
        ])

    def test_join_rename_join_key(self):
        df1 = pd.DataFrame([
            ['a', '2020-01-03', 1050],
            ['a', '2020-01-01', 1000],
            ['b', '2020-01-04', 990],
            ['a', '2020-01-02', 1100],
            ['b', '2020-01-03', 1200],
            ['c', '2020-01-07', 1250],
        ], columns=[
            'store',
            'date',
            'sold',
        ])
        df2 = pd.DataFrame([
            ['a', 'Store A', '2020-02-01'],
            ['b', 'Store B', '2020-02-02'],
        ], columns=[
            'store',
            'description',
            'date',
        ])
        base_action = BaseAction(dict(
            action_type='join',
            action_arguments=[100],
            action_code='',
            action_options=dict(
                left_on=['store'],
                right_on=['store'],
            ),
            action_variables=dict(),
            outputs=[
                {
                    'source_feature': {
                        'uuid': 'store',
                    },
                    'uuid': 'store_1',
                },
                {
                    'source_feature': {
                        'uuid': 'description',
                    },
                    'uuid': 'description',
                },
                {
                    'source_feature': {
                        'uuid': 'date',
                    },
                    'uuid': 'date_1',
                }
            ]
        ))
        df_new = base_action.execute(df1, df_to_join=df2)
        self.assertEqual(df_new.values.tolist(), [
            ['a', '2020-01-03', 1050, 'a', 'Store A', '2020-02-01'],
            ['a', '2020-01-01', 1000, 'a', 'Store A', '2020-02-01'],
            ['b', '2020-01-04', 990, 'b', 'Store B', '2020-02-02'],
            ['a', '2020-01-02', 1100, 'a', 'Store A', '2020-02-01'],
            ['b', '2020-01-03', 1200, 'b', 'Store B', '2020-02-02'],
            ['c', '2020-01-07', 1250, np.NaN, np.NaN, np.NaN],
        ])
        self.assertEqual(df_new.columns.to_list(), [
            'store', 'date', 'sold', 'store_1', 'description', 'date_1',
        ])

    def test_join_cast_to_str(self):
        df1 = pd.DataFrame([
            [1, '2020-01-03', 1050],
            [1, '2020-01-01', 1000],
            [2, '2020-01-04', 990],
            [1, '2020-01-02', 1100],
            [2, '2020-01-03', 1200],
            [3, '2020-01-07', 1250],
        ], columns=[
            'store',
            'date',
            'sold',
        ])
        df2 = pd.DataFrame([
            ['1', 'Store A'],
            ['2', 'Store B'],
        ], columns=[
            'store_name',
            'description',
        ])
        base_action = BaseAction(dict(
            action_type='join',
            action_arguments=[100],
            action_code='',
            action_options=dict(
                left_on=['store'],
                right_on=['store_name'],
                drop_columns=['store_name'],
                rename_columns={'description': 'store_description'}
            ),
            action_variables=dict(),
        ))
        df_new = base_action.execute(df1, df_to_join=df2)
        self.assertEqual(df_new.values.tolist(), [
            ['1', '2020-01-03', 1050, 'Store A'],
            ['1', '2020-01-01', 1000, 'Store A'],
            ['2', '2020-01-04', 990, 'Store B'],
            ['1', '2020-01-02', 1100, 'Store A'],
            ['2', '2020-01-03', 1200, 'Store B'],
            ['3', '2020-01-07', 1250, np.NaN],
        ])
        self.assertEqual(df_new.columns.to_list(), [
            'store', 'date', 'sold', 'store_description',
        ])



from unittest.mock import patch

from mage_ai.api.operations import constants
from mage_ai.tests.api.operations.test_base import BaseApiTestCase


class OperationTests(BaseApiTestCase):
    @patch('mage_ai.settings.DISABLE_NOTEBOOK_EDIT_ACCESS', 1)
    async def test_execute_create_with_disable_edit_access(self):
        operation = self.build_operation(
            action=constants.CREATE,
            payload=dict(block=dict(
                name='test block'
            )),
            resource='blocks',
        )
        response = await operation.execute()

        self.assertIsNotNone(response['error'])

    @patch('mage_ai.settings.DISABLE_NOTEBOOK_EDIT_ACCESS', 1)
    @patch('mage_ai.settings.REQUIRE_USER_AUTHENTICATION', 1)
    async def test_execute_create_with_disable_edit_access_and_user(self):
        operation = self.build_operation(
            action=constants.CREATE,
            payload=dict(block=dict(
                name='test block'
            )),
            resource='blocks',
        )
        response = await operation.execute()

        self.assertIsNotNone(response['error'])



from mage_ai.data_preparation.repo_manager import get_repo_config, get_repo_path
from mage_ai.server.constants import VERSION
import aiohttp


class Project():
    def __init__(self, repo_config=None):
        parts = get_repo_path().split('/')

        self.name = parts[-1]
        self.repo_config = repo_config or get_repo_config()
        self.version = VERSION

    @property
    def help_improve_mage(self) -> bool:
        return self.repo_config.help_improve_mage

    @property
    def project_uuid(self) -> str:
        return self.repo_config.project_uuid

    async def latest_version(self) -> str:
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    'https://pypi.org/pypi/mage-ai/json',
                    timeout=3,
                ) as response:
                    response_json = await response.json()
                    latest_version = response_json.get('info', {}).get('version', None)
        except Exception:
            latest_version = VERSION

        return latest_version



"""Create user roles, roles, and permissions

Revision ID: 83a74eca6ea1
Revises: 1bfc6d904929
Create Date: 2023-05-17 10:10:41.359684

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '83a74eca6ea1'
down_revision = '1bfc6d904929'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('role',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('permission',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('entity_id', sa.String(length=255), nullable=True),
    sa.Column('entity', sa.Enum('GLOBAL', 'PROJECT', 'PIPELINE', name='entity'), nullable=True),
    sa.Column('access', sa.Integer(), nullable=True),
    sa.Column('role_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['role_id'], ['role.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('user_role',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('role_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['role_id'], ['role.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('user_role')
    op.drop_table('permission')
    op.drop_table('role')
    # ### end Alembic commands ###



from mage_ai.data_preparation.repo_manager import RepoConfig
from mage_ai.services.spark.config import SparkConfig
import os


def get_spark_session(config: SparkConfig):
    from pyspark.conf import SparkConf
    from pyspark.sql import SparkSession

    repo_config = RepoConfig(repo_path=config.repo_path)
    spark_config = repo_config.spark_config

    if spark_config:
        conf = SparkConf()
        if spark_config.get('app_name'):
            conf.setAppName(spark_config.get('app_name'))
        if spark_config.get('spark_master'):
            conf.setMaster(spark_config.get('spark_master'))
        if spark_config.get('spark_home'):
            conf.setSparkHome(spark_config.get('spark_home'))
        if spark_config.get('executor_env'):
            list_kv_pairs = []
            for key, value in spark_config.get('executor_env').items():
                list_kv_pairs.append((key, value))
            conf.setExecutorEnv(key=None, value=None, pairs=list_kv_pairs)
        if spark_config.get('spark_jars'):
            conf.set('spark.jars', ','.join(spark_config.get('spark_jars')))
        if spark_config.get('others'):
            conf.setAll(spark_config.get('others'))

        return SparkSession.builder.config(conf=conf).getOrCreate()
    else:
        return SparkSession.builder.master(
            os.getenv('SPARK_MASTER_HOST', 'local')).getOrCreate()



from typing import Tuple
import os
import psutil


def get_compute() -> Tuple[float, float, float, float]:
    # Getting loadover15 minutes
    load1, load5, load15 = psutil.getloadavg()
    cpu_count = os.cpu_count()

    return load1, load5, load15, cpu_count


def get_memory() -> Tuple[float, float, float]:
    free_memory = None
    total_memory = None
    used_memory = None

    try:
        total_memory, used_memory, free_memory = map(
            int,
            os.popen('free -t -m').readlines()[-1].split()[1:],
        )
    except Exception as err:
        print(err)

    return free_memory, used_memory, total_memory



from dataclasses import dataclass
from google.cloud import storage
from google.oauth2 import service_account
from mage_ai.data_preparation.logging.logger_manager import LoggerManager
from mage_ai.data_preparation.repo_manager import RepoConfig
from mage_ai.shared.config import BaseConfig


@dataclass
class GCSConfig(BaseConfig):
    path_to_credentials: str
    bucket: str
    prefix: str


class GCSLoggerManager(LoggerManager):
    def __init__(
        self,
        repo_config: RepoConfig = None,
        **kwargs,
    ):
        super().__init__(repo_config=repo_config, **kwargs)
        self.gcs_config = GCSConfig.load(config=self.logging_config.destination_config)
        credentials = service_account.Credentials.from_service_account_file(
            self.gcs_config.path_to_credentials
        )
        self.gcs_client = storage.Client(credentials=credentials)

    def output_logs_to_destination(self):
        key = self.get_log_filepath()
        bucket = self.gcs_client.get_bucket(self.gcs_config.bucket)
        blob = bucket.blob(key)
        blob.upload_from_string(self.stream.getvalue())

    def create_log_filepath_dir(self, path):
        pass

    def get_log_filepath_prefix(self):
        return '{}/{}/{}/{}/{}'.format(
            self.gcs_config.prefix,
            self.repo_config.repo_name,
            'pipelines',
            self.pipeline_uuid,
            self.partition,
        )

    def get_logs(self):
        object_key = self.get_log_filepath()
        try:
            bucket = self.gcs_client.get_bucket(self.gcs_config.bucket)
            blob = bucket.blob(object_key)
            gcp_object = blob.download_as_string().decode('UTF-8')
        except Exception:
            gcp_object = ''

        return dict(
            name=object_key.split('/')[-1],
            path=object_key,
            content=gcp_object,
        )

    async def get_logs_async(self):
        """
        TODO: Implement this method
        """
        return self.get_logs()



from datetime import datetime
from mage_ai.api.resources.DatabaseResource import DatabaseResource
from mage_ai.orchestration.backfills.service import start_backfill, cancel_backfill
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import Backfill
from mage_ai.shared.hash import extract, merge_dict
from sqlalchemy import desc

ALLOWED_PAYLOAD_KEYS = [
    'block_uuid',
    'end_datetime',
    'interval_type',
    'interval_units',
    'name',
    'start_datetime',
]


class BackfillResource(DatabaseResource):
    datetime_keys = ['start_datetime', 'end_datetime']
    model_class = Backfill

    @classmethod
    @safe_db_query
    def collection(self, query_arg, meta, user, **kwargs):
        results = Backfill.query

        pipeline_uuid = query_arg.get('pipeline_uuid', [None])
        if pipeline_uuid:
            pipeline_uuid = pipeline_uuid[0]
        if pipeline_uuid:
            results = results.filter(Backfill.pipeline_uuid == pipeline_uuid)

        return results.order_by(desc(Backfill.created_at))

    @classmethod
    @safe_db_query
    def create(self, payload, user, **kwargs):
        pipeline_uuid = kwargs['parent_model'].uuid

        return super().create(merge_dict(
            extract(payload, ALLOWED_PAYLOAD_KEYS),
            dict(pipeline_uuid=pipeline_uuid),
        ), user, **kwargs)

    @safe_db_query
    def update(self, payload, **kwargs):
        pipeline_runs = []

        if 'status' in payload and payload['status'] != self.status:
            if Backfill.Status.INITIAL == payload['status']:
                pipeline_runs += start_backfill(self.model)
                return super().update(dict(
                    started_at=datetime.now(),
                    status=payload['status'],
                ))
            elif Backfill.Status.CANCELLED == payload['status']:
                cancel_backfill(self.model)
                return super().update(dict(status=payload['status']))

        return super().update(extract(payload, ALLOWED_PAYLOAD_KEYS))



from mage_ai.data_preparation.models.project import Project
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.oauth import User
from mage_ai.shared.environments import get_env
from mage_ai.shared.hash import merge_dict
from mage_ai.usage_statistics.constants import API_ENDPOINT, EventActionType, EventObjectType
from typing import Callable, Dict
import aiohttp
import json
import platform


class UsageStatisticLogger():
    def __init__(self, project=None):
        self.project = project or Project()

    @property
    def help_improve_mage(self) -> bool:
        return self.project.help_improve_mage

    async def project_impression(self) -> bool:
        if not self.help_improve_mage:
            return False

        return await self.__send_message(EventObjectType.PROJECT, EventActionType.IMPRESSION)

    @safe_db_query
    async def pipeline_runs_impression(self, count_func: Callable) -> bool:
        if not self.help_improve_mage:
            return False

        return await self.__send_message(
            EventObjectType.PIPELINE_RUN,
            EventActionType.IMPRESSION, dict(
                pipeline_runs=count_func(),
            ),
        )

    async def pipelines_impression(self, count_func: Callable) -> bool:
        if not self.help_improve_mage:
            return False

        return await self.__send_message(EventObjectType.PIPELINE, EventActionType.IMPRESSION, dict(
            pipelines=count_func(),
        ))

    @safe_db_query
    async def users_impression(self) -> bool:
        if not self.help_improve_mage:
            return False

        return await self.__send_message(EventObjectType.USER, EventActionType.IMPRESSION, dict(
            users=User.query.count(),
        ))

    async def __send_message(self, object_name: str, action_name: str, data: Dict = {}) -> bool:
        data_to_send = merge_dict(
            merge_dict(self.__shared_metadata(), dict(
                action=action_name,
                object=object_name,
            )),
            data,
        )

        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    API_ENDPOINT,
                    json=dict(
                        api_key='KwnbpKJNe6gOjC2X5ilxafFvxbNppiIfGejB2hlY',
                        usage_statistics=data_to_send,
                    ),
                    timeout=3,
                ) as response:
                    response_json = await response.json()
                    if response_json.get('success'):
                        print(json.dumps(data_to_send, indent=2))
                        return True
        except Exception as err:
            print(f'Error: {err}')

        return False

    def __shared_metadata(self) -> Dict:
        return dict(
            environment=get_env(),
            platform=platform.platform(),
            project_uuid=self.project.project_uuid,
            version=self.project.version,
        )



from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.export_utils import PandasTypes
from mage_ai.io.base import QUERY_ROW_LIMIT
from mage_ai.io.sql import BaseSQL
from pandas import DataFrame, Series
from typing import Any, IO, List, Union
import json
import numpy as np
import pyodbc


class MSSQL(BaseSQL):
    def __init__(
        self,
        database: str,
        host: str,
        password: str,
        user: str,
        schema: str = None,
        port: int = 1433,
        **kwargs,
    ) -> None:
        super().__init__(
            database=database,
            server=host,
            user=user,
            password=password,
            schema=schema,
            port=port,
            **kwargs
        )

    def _enforce_limit(self, query: str, limit: int = QUERY_ROW_LIMIT) -> str:
        # MSSQL doesn't support WITH statements in subqueries, so if the user uses a WITH
        # statement in their query, it would break if we tried to enforce a limit.
        return query

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'MSSQL':
        return cls(
            database=config[ConfigKey.MSSQL_DATABASE],
            schema=config[ConfigKey.MSSQL_SCHEMA],
            driver=config[ConfigKey.MSSQL_DRIVER],
            host=config[ConfigKey.MSSQL_HOST],
            password=config[ConfigKey.MSSQL_PASSWORD],
            port=config[ConfigKey.MSSQL_PORT],
            user=config[ConfigKey.MSSQL_USER],
        )

    def open(self) -> None:
        with self.printer.print_msg('Opening connection to MSSQL database'):
            driver = self.settings['driver']
            server = self.settings['server']
            database = self.settings['database']
            username = self.settings['user']
            password = self.settings['password']
            connection_string = (
                f'DRIVER={{{driver}}};'
                f'SERVER={server};'
                f'DATABASE={database};'
                f'UID={username};'
                f'PWD={password};'
                'ENCRYPT=yes;'
                'TrustServerCertificate=yes;'
            )
            self._ctx = pyodbc.connect(connection_string)

    def build_create_table_as_command(
        self,
        table_name: str,
        query_string: str
    ) -> str:
        return 'SELECT * INTO {}\nFROM ({}) AS prev'.format(
            table_name,
            query_string,
        )

    def table_exists(self, schema_name: str, table_name: str) -> bool:
        with self.conn.cursor() as cur:
            cur.execute('\n'.join([
                'SELECT TOP 1 * FROM information_schema.tables ',
                f'WHERE table_name = \'{table_name}\'',
            ]))
            return len(cur.fetchall()) >= 1

    def upload_dataframe(
        self,
        cursor: Any,
        df: DataFrame,
        db_dtypes: List[str],
        dtypes: List[str],
        full_table_name: str,
        buffer: Union[IO, None] = None,
        **kwargs,
    ) -> None:
        values_placeholder = ', '.join(["?" for i in range(len(df.columns))])
        values = []
        df_ = df.copy()
        columns = df_.columns
        for col in columns:
            dtype = df_[col].dtype
            if dtype == PandasTypes.OBJECT:
                df_[col] = df_[col].apply(lambda x: json.dumps(x))
            elif dtype in (
                PandasTypes.MIXED,
                PandasTypes.UNKNOWN_ARRAY,
                PandasTypes.COMPLEX,
            ):
                df_[col] = df_[col].astype('string')
        for i, row in df_.iterrows():
            values.append(tuple(row))

        sql = f'INSERT INTO {full_table_name} VALUES ({values_placeholder})'
        cursor.executemany(sql, values)

    def get_type(self, column: Series, dtype: str) -> str:
        if dtype in (
            PandasTypes.MIXED,
            PandasTypes.UNKNOWN_ARRAY,
            PandasTypes.COMPLEX,
            PandasTypes.OBJECT,
        ):
            return 'text'
        elif dtype in (PandasTypes.DATETIME, PandasTypes.DATETIME64):
            try:
                if column.dt.tz:
                    return 'datetime2'
            except AttributeError:
                pass
            return 'datetime2'
        elif dtype == PandasTypes.TIME:
            try:
                if column.dt.tz:
                    return 'time'
            except AttributeError:
                pass
            return 'time'
        elif dtype == PandasTypes.DATE:
            return 'date'
        elif dtype == PandasTypes.STRING:
            return 'char(255)'
        elif dtype == PandasTypes.CATEGORICAL:
            return 'text'
        elif dtype == PandasTypes.BYTES:
            return 'varbinary(255)'
        elif dtype in (PandasTypes.FLOATING, PandasTypes.DECIMAL, PandasTypes.MIXED_INTEGER_FLOAT):
            return 'decimal'
        elif dtype == PandasTypes.INTEGER:
            max_int, min_int = column.max(), column.min()
            if np.int16(max_int) == max_int and np.int16(min_int) == min_int:
                return 'bigint'
            elif np.int32(max_int) == max_int and np.int32(min_int) == min_int:
                return 'bigint'
            else:
                return 'bigint'
        elif dtype == PandasTypes.BOOLEAN:
            return 'char(52)'
        elif dtype in (PandasTypes.TIMEDELTA, PandasTypes.TIMEDELTA64, PandasTypes.PERIOD):
            return 'bigint'
        elif dtype == PandasTypes.EMPTY:
            return 'char(255)'
        else:
            print(f'Invalid datatype provided: {dtype}')

        return 'char(255)'



from collections import UserList
from mage_ai.api.api_context import ApiContext
from mage_ai.api.resources.Resource import Resource


class ResultSet(UserList):
    def __init__(self, results):
        self.context = ApiContext()
        self.data = []
        self.metadata = {}
        for r in results:
            if issubclass(r.__class__, Resource):
                key = r.__class__.__name__
                if not r.model_options.get('result_sets', None):
                    r.model_options['result_sets'] = {}

                existing_result_set = r.model_options['result_sets'].get(
                    key, None)
                if existing_result_set:
                    for i in self:
                        existing_result_set.append(i)
                    r.model_options['result_sets'][key] = existing_result_set
                else:
                    r.model_options['result_sets'][key] = self
            self.data.append(r)



from datetime import datetime
from json import JSONDecoder
import numpy as np

INTS = (
    np.int16,
    np.int32,
    np.int64,
    np.int8,
    np.int_,
    np.intc,
    np.intp,
    np.uint16,
    np.uint32,
    np.uint64,
    np.uint8,
)

MAX_ITEMS_IN_SAMPLE_OUTPUT = 20


def encode_complex(obj):
    if hasattr(obj, 'isoformat') and 'method' in type(obj.isoformat).__name__:
        return obj.isoformat()
    elif isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, INTS):
        return int(obj)
    elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
        return float(obj)
    elif isinstance(obj, (np.complex_, np.complex64, np.complex128)):
        return {'real': obj.real, 'imag': obj.imag}
    elif isinstance(obj, (np.ndarray,)):
        return obj.tolist()
    elif isinstance(obj, (np.bool_)):
        return bool(obj)
    elif isinstance(obj, (np.void)):
        return None

    return obj


def extract_json_objects(text, decoder=JSONDecoder()):
    """Find JSON objects in text, and yield the decoded JSON data

    Does not attempt to look for JSON arrays, text, or other JSON types outside
    of a parent JSON object.

    """
    pos = 0
    while True:
        match = text.find('{', pos)
        if match == -1:
            break
        try:
            result, index = decoder.raw_decode(text[match:])
            yield result
            pos = match + index
        except ValueError:
            pos = match + 1


def sample_output(obj):
    if isinstance(obj, list):
        sampled = len(obj) > MAX_ITEMS_IN_SAMPLE_OUTPUT
        return obj[:MAX_ITEMS_IN_SAMPLE_OUTPUT], sampled
    elif isinstance(obj, dict):
        sampled = False
        output = dict()
        for k, v in obj.items():
            v, v_sampled = sample_output(v)
            if v_sampled:
                sampled = True
            output[k] = v
        return output, sampled
    return obj, False



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Purpose

Shows how to use the AWS SDK for Python (Boto3) with the Amazon EMR API to create
and manage clusters and job steps.
"""

import logging

from botocore.exceptions import ClientError

logger = logging.getLogger(__name__)


def run_job_flow(
        name, log_uri, keep_alive, applications, job_flow_role, service_role,
        security_groups, steps, emr_client):
    """
    Runs a job flow with the specified steps. A job flow creates a cluster of
    instances and adds steps to be run on the cluster. Steps added to the cluster
    are run as soon as the cluster is ready.

    This example uses the 'emr-5.30.1' release. A list of recent releases can be
    found here:
        https://docs.aws.amazon.com/emr/latest/ReleaseGuide/emr-release-components.html.

    :param name: The name of the cluster.
    :param log_uri: The URI where logs are stored. This can be an Amazon S3 bucket URL,
                    such as 's3://my-log-bucket'.
    :param keep_alive: When True, the cluster is put into a Waiting state after all
                       steps are run. When False, the cluster terminates itself when
                       the step queue is empty.
    :param applications: The applications to install on each instance in the cluster,
                         such as Hive or Spark.
    :param job_flow_role: The IAM role assumed by the cluster.
    :param service_role: The IAM role assumed by the service.
    :param security_groups: The security groups to assign to the cluster instances.
                            Amazon EMR adds all needed rules to these groups, so
                            they can be empty if you require only the default rules.
    :param steps: The job flow steps to add to the cluster. These are run in order
                  when the cluster is ready.
    :param emr_client: The Boto3 EMR client object.
    :return: The ID of the newly created cluster.
    """
    try:
        response = emr_client.run_job_flow(
            Name=name,
            LogUri=log_uri,
            ReleaseLabel='emr-6.9.0',
            Instances={
                'MasterInstanceType': 'm5.xlarge',
                'SlaveInstanceType': 'm5.xlarge',
                'InstanceCount': 3,
                'KeepJobFlowAliveWhenNoSteps': keep_alive,
                'EmrManagedMasterSecurityGroup': security_groups['manager'].id,
                'EmrManagedSlaveSecurityGroup': security_groups['worker'].id,
            },
            Steps=[{
                'Name': step['name'],
                'ActionOnFailure': 'CONTINUE',
                'HadoopJarStep': {
                    'Jar': 'command-runner.jar',
                    'Args': ['spark-submit', '--deploy-mode', 'cluster',
                             step['script_uri'], *step['script_args']]
                }
            } for step in steps],
            Applications=[{
                'Name': app
            } for app in applications],
            JobFlowRole=job_flow_role.name,
            ServiceRole=service_role.name,
            EbsRootVolumeSize=10,
            VisibleToAllUsers=True
        )
        cluster_id = response['JobFlowId']
        logger.info("Created cluster %s.", cluster_id)
    except ClientError:
        logger.exception("Couldn't create cluster.")
        raise
    else:
        return cluster_id


def describe_cluster(cluster_id, emr_client):
    """
    Gets detailed information about a cluster.

    :param cluster_id: The ID of the cluster to describe.
    :param emr_client: The Boto3 EMR client object.
    :return: The retrieved cluster information.
    """
    try:
        response = emr_client.describe_cluster(ClusterId=cluster_id)
        cluster = response['Cluster']
        logger.info("Got data for cluster %s.", cluster['Name'])
    except ClientError:
        logger.exception("Couldn't get data for cluster %s.", cluster_id)
        raise
    else:
        return cluster


def terminate_cluster(cluster_id, emr_client):
    """
    Terminates a cluster. This terminates all instances in the cluster and cannot
    be undone. Any data not saved elsewhere, such as in an Amazon S3 bucket, is lost.

    :param cluster_id: The ID of the cluster to terminate.
    :param emr_client: The Boto3 EMR client object.
    """
    try:
        emr_client.terminate_job_flows(JobFlowIds=[cluster_id])
        logger.info("Terminated cluster %s.", cluster_id)
    except ClientError:
        logger.exception("Couldn't terminate cluster %s.", cluster_id)
        raise


def add_step(cluster_id, name, script_uri, script_args, emr_client):
    """
    Adds a job step to the specified cluster. This example adds a Spark
    step, which is run by the cluster as soon as it is added.

    :param cluster_id: The ID of the cluster.
    :param name: The name of the step.
    :param script_uri: The URI where the Python script is stored.
    :param script_args: Arguments to pass to the Python script.
    :param emr_client: The Boto3 EMR client object.
    :return: The ID of the newly added step.
    """
    try:
        response = emr_client.add_job_flow_steps(
            JobFlowId=cluster_id,
            Steps=[{
                'Name': name,
                'ActionOnFailure': 'CONTINUE',
                'HadoopJarStep': {
                    'Jar': 'command-runner.jar',
                    'Args': ['spark-submit', '--deploy-mode', 'cluster',
                             script_uri, *script_args]
                }
            }])
        step_id = response['StepIds'][0]
        logger.info("Started step with ID %s", step_id)
    except ClientError:
        logger.exception("Couldn't start step %s with URI %s.", name, script_uri)
        raise
    else:
        return step_id


def list_steps(cluster_id, emr_client):
    """
    Gets a list of steps for the specified cluster. In this example, all steps are
    returned, including completed and failed steps.

    :param cluster_id: The ID of the cluster.
    :param emr_client: The Boto3 EMR client object.
    :return: The list of steps for the specified cluster.
    """
    try:
        response = emr_client.list_steps(ClusterId=cluster_id)
        steps = response['Steps']
        logger.info("Got %s steps for cluster %s.", len(steps), cluster_id)
    except ClientError:
        logger.exception("Couldn't get steps for cluster %s.", cluster_id)
        raise
    else:
        return steps


def describe_step(cluster_id, step_id, emr_client):
    """
    Gets detailed information about the specified step, including the current state of
    the step.

    :param cluster_id: The ID of the cluster.
    :param step_id: The ID of the step.
    :param emr_client: The Boto3 EMR client object.
    :return: The retrieved information about the specified step.
    """
    try:
        response = emr_client.describe_step(ClusterId=cluster_id, StepId=step_id)
        step = response['Step']
        logger.info("Got data for step %s.", step_id)
    except ClientError:
        logger.exception("Couldn't get data for step %s.", step_id)
        raise
    else:
        return step



from mage_ai.io.base import BaseSQLDatabase, ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from pandas import DataFrame
from pyspark.sql import SparkSession
from typing import Dict, List, Union


class Spark(BaseSQLDatabase):
    """
    Handles data transfer between a Spark session and the Mage app.
    """
    def _get_spark_session(self, **kwargs):
        if self.spark_init:
            return self.spark
        try:
            self.spark = SparkSession.builder.master(
                kwargs.get('host', 'local')).getOrCreate()
            self.spark_init = True
        except Exception:
            self.spark = None
        return self.spark

    def __init__(self, **kwargs) -> None:
        """
        Initializes settings for connecting to a Spark session.
        """
        if kwargs.get('verbose') is not None:
            kwargs.pop('verbose')
        super().__init__(verbose=kwargs.get('verbose', True))
        self.spark = None
        self.spark_init = False
        with self.printer.print_msg('Connecting to Spark Session'):
            self.client = self._get_spark_session(**kwargs)

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'Spark':
        """
        Initializes Spark Session client from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        if ConfigKey.SPARK_HOST not in config:
            raise ValueError(
                'No valid configuration settings found for Spark Session. '
                'You must specify host.'
            )
        return cls(
            method=config[ConfigKey.SPARK_METHOD],
            host=config[ConfigKey.SPARK_HOST],
            database=config[ConfigKey.SPARK_SCHEMA],
        )

    def execute(self, query_string: str, **kwargs) -> None:
        """
        Sends query to the connected Spark Session.

        Args:
            query_string (str): Query to execute on the Spark Session.
            **kwargs: Additional arguments to pass to query, such as query configurations
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            self.client.sql(query_string, **kwargs)

    def execute_query(
        self,
        query: str,
        parameters: Dict = None,
        **kwargs,
    ) -> DataFrame:
        """
        Sends query to the connected Spark Session.

        Args:
            query (str): Query to execute on the Spark Session.
            **kwargs: Additional arguments to pass to query, such as query configurations
        """
        query = self._clean_query(query)
        with self.printer.print_msg(f'Executing query \'{query}\''):
            result = self.client.sql(query, parameters=parameters).toPandas()

        return result

    def execute_queries(
        self,
        queries: List[str],
        query_variables: List[Dict] = None,
        fetch_query_at_indexes: List[bool] = None,
        **kwargs,
    ) -> List:
        results = []

        for idx, query in enumerate(queries):
            parameters = (query_variables[idx]
                          if query_variables and idx < len(query_variables)
                          else {})
            query = self._clean_query(query)
            result = self.client.sql(query, parameters=parameters)

            if (fetch_query_at_indexes and idx < len(fetch_query_at_indexes) and
                    fetch_query_at_indexes[idx]):
                result = result.toPandas()

            results.append(result)

        return results

    def load(
        self,
        query_string: str,
        limit: int = QUERY_ROW_LIMIT,
        display_query: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from Spark Session into a Pandas data frame based on the query given.
        This will fail if the query returns no data from the database. When a select query
        is provided, this function will load at maximum 10,000,000 rows of data. To operate on more
        data, consider performing data transformations in warehouse.

        Args:
            query_string (str): Query to fetch a table or subset of a table.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to
                                    10,000,000.
            **kwargs: Additional arguments to pass to query, such as query configurations

        Returns:
            DataFrame: Data frame associated with the given query.
        """

        print_message = 'Loading data'
        if verbose:
            print_message += ' with query'

            if display_query:
                for line in display_query.split('\n'):
                    print_message += f'\n{line}'
            else:
                print_message += f'\n{query_string}'

        query_string = self._clean_query(query_string)

        with self.printer.print_msg(print_message):
            return self.client.sql(
                self._enforce_limit(query_string, limit), **kwargs
            ).toPandas()

    def export(
        self,
        df: DataFrame,
        table_name: str,
        database: str = 'default',
        if_exists: str = 'append',
        index: bool = False,
        query_string: Union[str, None] = None,
        create_table_statement: Union[str, None] = None,
        verbose: bool = True,
        **kwargs,
    ) -> None:
        """
        Exports a Pandas data frame to a Spark Session based on the table name.
        If table doesn't exist, the table is automatically created.

        Args:
            df (DataFrame): Data frame to export
            table_name (str): Name of the table to export data to (excluding database).
            If this table exists, the table schema must match the data frame schema.
            If this table doesn't exist, query_string must be specified to create the new table.
            database (str): Name of the database in which the table is located.
            if_exists (str, optional): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table. In this case the schema must
                                match the original table.
            Defaults to `'append'`.
            **kwargs: Additional arguments to pass to writer
        """

        if type(df) is dict:
            df = DataFrame([df])
        elif type(df) is list:
            df = DataFrame(df)

        def __process(database: Union[str, None]):
            table_exists = self.client.catalog.tableExists(f'{database}.{table_name}')
            should_create_table = not table_exists

            if table_exists:
                if ExportWritePolicy.FAIL == if_exists:
                    raise ValueError(
                        f'Table \'{table_name}\' already'
                        ' exists in database {database}.',
                    )
                elif ExportWritePolicy.REPLACE == if_exists:
                    self.client.sql(
                        f'DROP TABLE IF EXISTS {database}.{table_name}')
                    should_create_table = True

            if query_string:
                self.client.sql(f'USE {database}')

                if should_create_table:
                    self.client.sql(f"""
CREATE TABLE IF NOT EXISTS {database}.{table_name} AS
{query_string}
""")
                else:
                    self.client.sql(f"""
INSERT INTO {database}.{table_name}
{query_string}
""")
            else:
                if should_create_table:
                    self.client.sql(create_table_statement)
                spark_df = self.client.createDataFrame(df)
                spark_df.write.mode('append').saveAsTable(f'{database}.{table_name}')

        if verbose:
            with self.printer.print_msg(
                    f'Exporting data to table \'{database}.{table_name}\''):
                __process(database=database)
        else:
            __process(database=database)



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig


@dataclass
class AirbyteConfig(BaseConfig):
    api_version: str = 'v1'
    host: str = 'localhost'
    password: str = 'password'
    port: int = 8000
    use_ssl: bool = False
    username: str = 'airbyte'



from dataclasses import dataclass
from mage_ai.shared.config import BaseConfig
from mage_ai.streaming.sinks.base import BaseSink
from opensearchpy import OpenSearch, RequestsHttpConnection
from opensearchpy.helpers import bulk
from requests_aws4auth import AWS4Auth
from typing import Dict, List
import boto3
import time


@dataclass
class OpensearchConfig(BaseConfig):
    host: str
    index_name: str
    verify_certs: bool = True
    http_auth: str = '@awsauth'


class OpenSearchSink(BaseSink):
    config_class = OpensearchConfig

    def init_client(self):
        # Initialize opensearch client
        if self.config.http_auth == '@awsauth':
            session = boto3.Session()
            credentials = session.get_credentials()
            awsauth = AWS4Auth(
              credentials.access_key,
              credentials.secret_key,
              session.region_name,
              'es',
              session_token=credentials.token,
            )
            http_auth = awsauth
        else:
            http_auth = self.config.http_auth

        self.client = OpenSearch(
                hosts=[self.config.host],
                http_auth=http_auth,
                use_ssl=True,
                verify_certs=self.config.verify_certs,
                ssl_assert_hostname=False,
                ssl_show_warn=False,
                connection_class=RequestsHttpConnection,
        )

    def test_connection(self):
        return True

    def write(self, data: Dict):
        self._print(f'[Opensearch] Ingest data {data}, time={time.time()}')
        self.client.index(
            index=self.config.index_name,
            body=data,
            refresh=True
        )

    def batch_write(self, data: List[Dict]):
        self._print(f'[Opensearch] Batch ingest data {data}, time={time.time()}')
        docs = [{'_index': self.config.index_name, '_source': doc} for doc in data]
        bulk(self.client, docs)



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.WorkspacePresenter import WorkspacePresenter
from mage_ai.orchestration.db.models.oauth import Permission


class WorkspacePolicy(BasePolicy):
    @property
    def entity(self):
        return Permission.Entity.GLOBAL, None


WorkspacePolicy.allow_actions([
    constants.DETAIL,
    constants.LIST,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_admin_role())

WorkspacePolicy.allow_actions([
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.is_owner())

WorkspacePolicy.allow_read(WorkspacePresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
    constants.LIST,
], condition=lambda policy: policy.has_at_least_admin_role())

WorkspacePolicy.allow_read(WorkspacePresenter.default_attributes + [
    'success',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.DELETE,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_admin_role())

WorkspacePolicy.allow_write([
    'cluster_type',
    'name',
    'namespace',
    'storage_class_name',
    'service_account_name',
    'container_config',
    'cluster_name',
    'task_definition',
    'container_name',
    'project_id',
    'path_to_credentials',
    'region',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.CREATE,
    constants.UPDATE,
], condition=lambda policy: policy.is_owner())

WorkspacePolicy.allow_query([
    'cluster_type',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
    constants.LIST,
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_admin_role())



import argparse
import asyncio
import os
import traceback
import webbrowser
from time import sleep
from typing import Union

import tornado.ioloop
import tornado.web
from tornado import autoreload
from tornado.ioloop import PeriodicCallback
from tornado.log import enable_pretty_logging
from tornado.options import options

from mage_ai.authentication.passwords import create_bcrypt_hash, generate_salt
from mage_ai.data_preparation.repo_manager import (
    ProjectType,
    get_project_type,
    init_repo,
    set_repo_path,
)
from mage_ai.data_preparation.shared.constants import MANAGE_ENV_VAR
from mage_ai.orchestration.db import db_connection
from mage_ai.orchestration.db.database_manager import database_manager
from mage_ai.orchestration.db.models.oauth import Oauth2Application, Role, User
from mage_ai.server.active_kernel import switch_active_kernel
from mage_ai.server.api.base import BaseHandler
from mage_ai.server.api.blocks import ApiPipelineBlockAnalysisHandler
from mage_ai.server.api.events import (
    ApiEventHandler,
    ApiEventMatcherDetailHandler,
    ApiEventMatcherListHandler,
)
from mage_ai.server.api.triggers import ApiTriggerPipelineHandler
from mage_ai.server.api.v1 import (
    ApiChildDetailHandler,
    ApiChildListHandler,
    ApiResourceDetailHandler,
    ApiResourceListHandler,
)
from mage_ai.server.constants import DATA_PREP_SERVER_PORT
from mage_ai.server.docs_server import run_docs_server
from mage_ai.server.kernel_output_parser import parse_output_message
from mage_ai.server.kernels import DEFAULT_KERNEL_NAME
from mage_ai.server.scheduler_manager import (
    SCHEDULER_AUTO_RESTART_INTERVAL,
    check_scheduler_status,
    scheduler_manager,
)
from mage_ai.server.subscriber import get_messages
from mage_ai.server.terminal_server import (
    MageTermManager,
    MageUniqueTermManager,
    TerminalWebsocketServer,
)
from mage_ai.server.websocket_server import WebSocketServer
from mage_ai.settings import (
    AUTHENTICATION_MODE,
    LDAP_ADMIN_USERNAME,
    OAUTH2_APPLICATION_CLIENT_ID,
    REQUIRE_USER_AUTHENTICATION,
    SERVER_VERBOSITY,
    SHELL_COMMAND,
    USE_UNIQUE_TERMINAL,
)
from mage_ai.shared.logger import LoggingLevel
from mage_ai.shared.utils import is_port_in_use
from mage_ai.usage_statistics.logger import UsageStatisticLogger


class MainPageHandler(tornado.web.RequestHandler):
    def get(self, *args):
        self.render('index.html')


class PipelineRunsPageHandler(tornado.web.RequestHandler):
    def get(self, *args):
        self.render('pipeline-runs.html')


class ManagePageHandler(tornado.web.RequestHandler):
    def get(self, *args):
        self.render('manage.html')


class ApiSchedulerHandler(BaseHandler):
    def get(self, action_type=None):
        self.write(dict(scheduler=dict(status=scheduler_manager.get_status())))

    def post(self, action_type):
        if action_type == 'start':
            scheduler_manager.start_scheduler()
        elif action_type == 'stop':
            scheduler_manager.stop_scheduler()
        self.write(dict(scheduler=dict(status=scheduler_manager.get_status())))


def make_app():
    shell_command = SHELL_COMMAND
    if shell_command is None:
        shell_command = 'bash'
        if os.name == 'nt':
            shell_command = 'cmd'
    term_klass = MageTermManager
    if USE_UNIQUE_TERMINAL:
        term_klass = MageUniqueTermManager
    term_manager = term_klass(shell_command=[shell_command])

    routes = [
        (r'/', MainPageHandler),
        (r'/pipelines', MainPageHandler),
        (r'/pipelines/(.*)', MainPageHandler),
        (r'/pipeline-runs', PipelineRunsPageHandler),
        (r'/settings', MainPageHandler),
        (r'/settings/(.*)', MainPageHandler),
        (r'/sign-in', MainPageHandler),
        (r'/terminal', MainPageHandler),
        (r'/triggers', MainPageHandler),
        (r'/manage', ManagePageHandler),
        (
            r'/_next/static/(.*)',
            tornado.web.StaticFileHandler,
            {'path': os.path.join(os.path.dirname(__file__), 'frontend_dist/_next/static')},
        ),
        (
            r'/fonts/(.*)',
            tornado.web.StaticFileHandler,
            {'path': os.path.join(os.path.dirname(__file__), 'frontend_dist/fonts')},
        ),
        (
            r'/images/(.*)',
            tornado.web.StaticFileHandler,
            {'path': os.path.join(os.path.dirname(__file__), 'frontend_dist/images')},
        ),
        (
            r'/(favicon.ico)',
            tornado.web.StaticFileHandler,
            {'path': os.path.join(os.path.dirname(__file__), 'frontend_dist')},
        ),
        (r'/websocket/', WebSocketServer),
        (r'/websocket/terminal', TerminalWebsocketServer, {'term_manager': term_manager}),
        # Not sure what is using this, perhaps the event triggering via Lambda?
        (r'/api/events', ApiEventHandler),
        (r'/api/event_matchers', ApiEventMatcherListHandler),
        (r'/api/event_matchers/(?P<event_matcher_id>\w+)', ApiEventMatcherDetailHandler),
        # TODO: This call is not easily removed from the frontend so will change this
        # in a future PR.
        (
            r'/api/pipelines/(?P<pipeline_uuid>\w+)/blocks/(?P<block_uuid>[\w\%2f\.]+)/analyses',
            ApiPipelineBlockAnalysisHandler,
        ),

        # Trigger pipeline via API
        (
            r'/api/pipeline_schedules/(?P<pipeline_schedule_id>\w+)/pipeline_runs/(?P<token>\w+)',
            ApiTriggerPipelineHandler,
        ),

        # API v1 routes
        (
            r'/api/(?P<resource>\w+)/(?P<pk>[\w\%2f\.]+)/(?P<child>\w+)/(?P<child_pk>[\w\%2f\.]+)',
            ApiChildDetailHandler,
        ),
        (
            r'/api/(?P<resource>\w+)/(?P<pk>[\w\%2f\.]+)/(?P<child>\w+)',
            ApiChildListHandler,
        ),
        (
            r'/api/(?P<resource>\w+)/(?P<pk>[\w\%2f\.]+)',
            ApiResourceDetailHandler,
        ),
        (r'/api/(?P<resource>\w+)', ApiResourceListHandler),
        (r'/api/(?P<resource>\w+)/(?P<pk>.+)', ApiResourceDetailHandler),
    ]
    autoreload.add_reload_hook(scheduler_manager.stop_scheduler)
    return tornado.web.Application(
        routes,
        autoreload=True,
        template_path=os.path.join(os.path.dirname(__file__), 'frontend_dist'),
    )


async def main(
    host: Union[str, None] = None,
    port: Union[str, None] = None,
    project: Union[str, None] = None,
):
    switch_active_kernel(DEFAULT_KERNEL_NAME)

    app = make_app()

    port = int(port)
    max_port = port + 100
    while is_port_in_use(port):
        if port > max_port:
            raise Exception(
                'Unable to find an open port, please clear your running processes if possible.'
            )
        port += 1

    app.listen(
        port,
        address=host if host != 'localhost' else None,
    )

    url = f'http://{host or "localhost"}:{port}'
    webbrowser.open_new_tab(url)
    print(f'Mage is running at {url} and serving project {project}')

    db_connection.start_session(force=True)

    if REQUIRE_USER_AUTHENTICATION:
        print('User authentication is enabled.')
        # We need to sleep for a few seconds after creating all the tables or else there
        # may be an error trying to create users.
        sleep(3)

        # Create new roles on existing users. This should only need to be run once.
        Role.create_default_roles()

        # Fetch legacy owner user to check if we need to batch update the users with new roles.
        legacy_owner_user = User.query.filter(User._owner == True).first()  # noqa: E712

        default_owner_role = Role.get_role('Owner')
        owner_users = default_owner_role.users if default_owner_role else []
        if not legacy_owner_user and len(owner_users) == 0:
            print('User with owner permission doesn’t exist, creating owner user.')
            if AUTHENTICATION_MODE.lower() == 'ldap':
                user = User.create(
                    roles_new=[default_owner_role],
                    username=LDAP_ADMIN_USERNAME,
                )
            else:
                password_salt = generate_salt()
                user = User.create(
                    email='admin@admin.com',
                    password_hash=create_bcrypt_hash('admin', password_salt),
                    password_salt=password_salt,
                    roles_new=[default_owner_role],
                    username='admin',
                )
            owner_user = user
        else:
            if legacy_owner_user and not legacy_owner_user.roles_new:
                User.batch_update_user_roles()
            owner_user = next(iter(owner_users), None) or legacy_owner_user

        oauth_client = Oauth2Application.query.filter(
            Oauth2Application.client_id == OAUTH2_APPLICATION_CLIENT_ID,
        ).first()
        if not oauth_client:
            print('OAuth2 application doesn’t exist for frontend, creating OAuth2 application.')
            oauth_client = Oauth2Application.create(
                client_id=OAUTH2_APPLICATION_CLIENT_ID,
                client_type=Oauth2Application.ClientType.PUBLIC,
                name='frontend',
                user_id=owner_user.id,
            )

    # Check scheduler status periodically
    periodic_callback = PeriodicCallback(
        check_scheduler_status,
        SCHEDULER_AUTO_RESTART_INTERVAL,
    )
    periodic_callback.start()

    get_messages(
        lambda content: WebSocketServer.send_message(
            parse_output_message(content),
        ),
    )

    await asyncio.Event().wait()


def start_server(
    host: str = None,
    port: str = None,
    project: str = None,
    manage: bool = False,
    dbt_docs: bool = False,
):
    host = host if host else None
    port = port if port else DATA_PREP_SERVER_PORT
    project = project if project else None

    # Set project path in environment variable
    if project:
        project = project = os.path.abspath(project)
    else:
        project = os.path.join(os.getcwd(), 'default_repo')

    if not os.path.exists(project):
        init_repo(project)
    set_repo_path(project)

    asyncio.run(UsageStatisticLogger().project_impression())

    if dbt_docs:
        run_docs_server()
    else:
        project_type = get_project_type()
        if manage or project_type == ProjectType.MAIN:
            os.environ[MANAGE_ENV_VAR] = '1'
            # run migrations to enable user authentication
            try:
                database_manager.run_migrations()
            except Exception:
                traceback.print_exc()
        else:
            # Start a subprocess for scheduler
            scheduler_manager.start_scheduler()

        if LoggingLevel.is_valid_level(SERVER_VERBOSITY):
            options.logging = SERVER_VERBOSITY
        enable_pretty_logging()

        # Start web server
        asyncio.run(
            main(
                host=host,
                port=port,
                project=project,
            )
        )


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, default=None)
    parser.add_argument('--port', type=str, default=None)
    parser.add_argument('--project', type=str, default=None)
    parser.add_argument('--manage-instance', type=str, default='0')
    parser.add_argument('--dbt-docs-instance', type=str, default='0')
    args = parser.parse_args()

    host = args.host
    port = args.port
    project = args.project
    manage = args.manage_instance == '1'
    dbt_docs = args.dbt_docs_instance == '1'

    start_server(
        host=host,
        port=port,
        project=project,
        manage=manage,
        dbt_docs=dbt_docs,
    )



from mage_ai.orchestration.job_manager import JobManager, JobType
from mage_ai.orchestration.queue.queue_factory import QueueFactory
from mage_ai.tests.base_test import TestCase
from unittest.mock import MagicMock


class JobMnagerTests(TestCase):
    def setUp(self):
        self.mock_queue = MagicMock()
        QueueFactory.get_queue = MagicMock(return_value=self.mock_queue)
        self.job_manager = JobManager()

    def test_add_job(self):
        def test_func(arg1: int, arg2: int):
            return arg1 + arg2

        self.job_manager.add_job(JobType.PIPELINE_RUN, 1, test_func, 1, 2)

        self.mock_queue.enqueue.assert_called_once_with(
            'pipeline_run_1',
            test_func,
            1,
            2,
        )

    def test_clean_up_jobs(self):
        self.job_manager.clean_up_jobs()
        self.mock_queue.clean_up_jobs.assert_called_once()

    def test_has_block_run_job(self):
        self.job_manager.has_block_run_job(2)
        self.mock_queue.has_job.assert_called_once_with('block_run_2')

    def test_has_pipeline_run_job(self):
        self.job_manager.has_pipeline_run_job(3)
        self.mock_queue.has_job.assert_called_once_with('pipeline_run_3')

    def test_kill_block_run_job(self):
        self.job_manager.kill_block_run_job(4)
        self.mock_queue.kill_job.assert_called_once_with('block_run_4')

    def test_kill_pipeline_run_job(self):
        self.job_manager.kill_pipeline_run_job(5)
        self.mock_queue.kill_job.assert_called_once_with('pipeline_run_5')



import datetime
import pytz


def compare(date1: datetime, date2: datetime) -> int:
    if date1 is None or date2 is None:
        return None

    date1_utc = date1.astimezone(pytz.UTC)
    date2_utc = date2.astimezone(pytz.UTC)

    if date1_utc < date2_utc:
        return -1
    elif date1_utc > date2_utc:
        return 1
    else:
        return 0


def n_days_ago(n: int) -> str:
    day = datetime.datetime.now() - datetime.timedelta(days=n)
    return day.strftime('%Y-%m-%d')


def str_to_timedelta(period_str: str):
    unit = period_str[-1]
    if unit not in ['d', 'h', 'w']:
        raise Exception(
            'Please provide a valid period unit ("d", "h", or "w")')
    if unit == 'd':
        return datetime.timedelta(days=int(period_str[:-1]))
    elif unit == 'h':
        return datetime.timedelta(hours=int(period_str[:-1]))
    elif unit == 'w':
        return datetime.timedelta(weeks=int(period_str[:-1]))
    return None



{% extends "data_loaders/default.jinja" %}
{% block imports %}
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.mongodb import MongoDB
from os import path
{{ super() -}}
{% endblock %}


{% block content %}
@data_loader
def load_from_mongodb(*args, **kwargs):
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    query = {}

    return MongoDB.with_config(ConfigFileLoader(config_path, config_profile)).load(
        query=query,
        collection='collection_name',
    )
{% endblock %}



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.IntegrationSourceStreamPresenter import IntegrationSourceStreamPresenter


class IntegrationSourceStreamPolicy(BasePolicy):
    pass


IntegrationSourceStreamPolicy.allow_actions([
    constants.UPDATE,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

IntegrationSourceStreamPolicy.allow_read(IntegrationSourceStreamPresenter.default_attributes + [
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.UPDATE,
], condition=lambda policy: policy.has_at_least_viewer_role())



from mage_ai.data_preparation.models.block.sql.utils.shared import (
    create_upstream_block_tables as create_upstream_block_tables_orig,
    interpolate_input,
)
from mage_ai.io.config import ConfigKey
from typing import Dict


def create_upstream_block_tables(
    loader,
    block,
    cascade_on_drop: bool = False,
    configuration: Dict = None,
    execution_partition: str = None,
    cache_upstream_dbt_models: bool = False,
    query: str = None,
):
    create_upstream_block_tables_orig(
        loader,
        block,
        cascade_on_drop,
        configuration,
        execution_partition,
        cache_upstream_dbt_models,
        cache_keys=[
            ConfigKey.MSSQL_DATABASE,
            ConfigKey.MSSQL_SCHEMA,
            ConfigKey.MSSQL_HOST,
            ConfigKey.MSSQL_PORT,
        ],
        no_schema=True,
        query=query,
    )


def interpolate_input_data(block, query):
    return interpolate_input(
        block,
        query,
        lambda db, schema, tn: tn,
    )



from mage_ai.data_cleaner.column_types.constants import NUMBER_TYPES
from mage_ai.data_cleaner.transformer_actions.action_code import query_with_action_code
from mage_ai.data_cleaner.transformer_actions.custom_action import execute_custom_action
import pandas as pd


def custom(df, action, **kwargs):
    return execute_custom_action(df, action, **kwargs)


def drop_duplicates(df, action, **kwargs):
    keep = action.get('action_options', {}).get('keep', 'last')
    action_args = dict(keep=keep)
    subset_cols = action.get('action_arguments')
    if subset_cols is not None and len(subset_cols) > 0:
        action_args['subset'] = subset_cols
    return df.drop_duplicates(**action_args)


def filter_rows(df, action, **kwargs):
    """
    df:
        Pandas DataFrame
    action:
        TransformerAction serialized into a dictionary
    """
    action_code = action['action_code']

    return query_with_action_code(df, action_code, kwargs)


def sort_rows(df, action, **kwargs):
    ascending = action.get('action_options', {}).get('ascending', True)
    ascendings = action.get('action_options', {}).get('ascendings', [])
    if len(ascendings) > 0:
        ascending = ascendings[0]

    feature_by_uuid = {}
    if action.get('action_variables'):
        for _, val in action['action_variables'].items():
            feature = val.get('feature')
            if feature:
                feature_by_uuid[feature['uuid']] = feature

    na_indexes = None
    as_types = {}

    for idx, uuid in enumerate(action['action_arguments']):
        feature = feature_by_uuid.get(uuid)
        if feature and feature['column_type'] in NUMBER_TYPES:
            as_types[uuid] = float
            if idx == 0:
                na_indexes = df[(df[uuid].isnull()) | (df[uuid].astype(str).str.len() == 0)].index

    bad_df = None
    if na_indexes is not None:
        bad_df = df.index.isin(na_indexes)

    index = (
        (df[~bad_df] if bad_df is not None else df)
        .astype(as_types)
        .sort_values(
            by=action['action_arguments'],
            ascending=ascendings if len(ascendings) > 0 else ascending,
        )
        .index
    )

    df_final = df.loc[index]
    if bad_df is not None:
        if ascending:
            return pd.concat(
                [
                    df.iloc[bad_df],
                    df_final,
                ]
            )

        return pd.concat(
            [
                df_final,
                df.iloc[bad_df],
            ]
        )

    return df_final


def remove_row(df, action, **kwargs):
    row_indices = action['action_options']['rows']
    row_indices = [index for index in row_indices if index in df.index]
    return df.drop(row_indices, axis=0)
